from .mdf3 import MDF3
from .mdf4 import MDF4
from .mdf import MDF
from .signal import Signal

__version__ = '2.5.1'

__all__ = ['__version__',
           'MDF',
           'MDF3',
           'MDF4',
           'Signal']



"""
MDF file format module

"""
import csv
import os
from warnings import warn

import numpy as np

from .mdf3 import MDF3
from .mdf4 import MDF4
from .v3constants import CHANNEL_TYPE_MASTER as V3_MASTER
from .v4constants import CHANNEL_TYPE_MASTER as V4_MASTER
from .v4constants import CHANNEL_TYPE_VIRTUAL_MASTER as V4_VIRTUAL_MASTER
from .utils import MdfException


MDF3_VERSIONS = ('3.00', '3.10', '3.20', '3.30')
MDF4_VERSIONS = ('4.00', '4.10', '4.11')


__all__ = ['MDF', ]


class MDF(object):
    """Unified access to MDF v3 and v4 files.

    Parameters
    ----------
    name : string
        mdf file name, if provided it must be a real file name
    load_measured_data : bool
        load data option; default *True*

        * if *True* the data group binary data block will be loaded in RAM
        * if *False* the channel data is read from disk on request
    version : string
        mdf file version ('3.00', '3.10', '3.20', '3.30', '4.00', '4.10', '4.11'); default '3.20'

    """
    def __init__(self, name=None, load_measured_data=True, version='3.20'):
        if name:
            if os.path.isfile(name):
                with open(name, 'rb') as file_stream:
                    file_stream.read(8)
                    version = file_stream.read(4).decode('ascii')
                if version in MDF3_VERSIONS:
                    self._file = MDF3(name, load_measured_data)
                elif version in MDF4_VERSIONS:
                    self._file = MDF4(name, load_measured_data)
            else:
                raise MdfException('File "{}" does not exist'.format(name))
        else:
            if version in MDF3_VERSIONS:
                self._file = MDF3(version=version, load_measured_data=load_measured_data)
            elif version in MDF4_VERSIONS:
                self._file = MDF4(version=version, load_measured_data=load_measured_data)

        # link underlying _file attributes and methods to the new MDF object
        for attr in set(dir(self._file)) - set(dir(self)):
            setattr(self, attr, getattr(self._file, attr))

    def convert(self, to, load_measured_data=True):
        """convert MDF to other versions

        Parameters
        ----------
        to : str
            new mdf version from ('3.00', '3.10', '3.20', '3.30', '4.00', '4.10', '4.11')
        load_measured_data : bool
            load data option; default *True*

            * if *True* the data group binary data block will be loaded in RAM
            * if *False* the channel data is stored to a temporary file and read from disk on request

        Returns
        -------
        out : MDF
            new MDF object

        """
        if not to in MDF3_VERSIONS + MDF4_VERSIONS:
            warn('Unknown output mdf version "{}". Available versions are {}'.format(to, MDF4_VERSIONS + MDF3_VERSIONS))
            return
        else:
            out = MDF(version=to, load_measured_data=load_measured_data)
            if self.version in MDF3_VERSIONS:
                master_type = (V3_MASTER,)
            else:
                master_type = (V4_MASTER, V4_VIRTUAL_MASTER)

            # walk through all groups and get all channels
            for i, gp in enumerate(self.groups):
                sigs = []
                for j, ch in enumerate(gp['channels']):
                    if not ch['channel_type'] in master_type:
                        sigs.append(self.get(group=i, index=j))
                out.append(sigs,
                           'Converted from {} to {}'.format(self.version, to),
                           common_timebase=True)
            return out

    def cut(self, start=None, stop=None):
        """convert MDF to other versions

        Parameters
        ----------
        start : float
            start time, default None. If *None* then the start of measurement is used
        stop : float
            stop time, default . If *None* then the end of measurement is used

        Returns
        -------
        out : MDF
            new MDF object

        """
        out = MDF(version=self.version)
        if self.version in MDF3_VERSIONS:
            master_type = (V3_MASTER,)
        else:
            master_type = (V4_MASTER, V4_VIRTUAL_MASTER)

        # walk through all groups and get all channels
        for i, gp in enumerate(self.groups):
            sigs = []
            for j, ch in enumerate(gp['channels']):
                if not ch['channel_type'] in master_type:
                    sigs.append(self.get(group=i, index=j).cut(start, stop))
            out.append(sigs,
                       'Cut from {} to {}'.format('{}s'.format(start) if start else 'start of measurement',
                                                  '{}s'.format(stop) if stop else 'end of measurement'),
                       common_timebase=True)
        return out

    def export(self, format, filename=None):
        """ export MDF to other formats. The *MDF* file name is used is available,
        else the *filename* aragument must be provided.

        Parameters
        ----------
        format : string
            can be one of the following:

                * *csv* : CSV export that uses the ";" delimiter. This option wil generate a new csv file for each data group (<MDFNAME>_DataGroup_XX.csv).
                * *hdf5* : HDF5 file output; each *MDF* data group is mapped to a *HDF5* group with the name 'DataGroup_xx' (where xx is the index)
                * *excel* : Excel file output (very slow). This option wil generate a new excel file for each data group (<MDFNAME>_DataGroup_XX.xlsx).

        filename : string
            export file name

        """
        if filename is None and self.name is None:
            warn('Must specify filename for export if MDF was created without a file name')
        else:
            name = self.name if self.name else filename
            if format == 'hdf5':
                try:
                    from h5py import File as HDF5
                except ImportError:
                    warn('h5py not found; export to HDF5 is unavailable')
                    return
                else:
                    name = os.path.splitext(name)[0] + '.hdf'
                    with HDF5(name, 'w') as f:
                        # header information
                        group = f.create_group(os.path.basename(name))

                        if self.version in MDF3_VERSIONS:
                            for item in ('date', 'time', 'author', 'organization', 'project', 'subject'):
                                group.attrs[item] = self.header[item]

                        # save each data group in a HDF5 group called "DataGroup_xx"
                        # with the index starting from 1
                        # each HDF5 group will have a string attribute "master" that
                        # will hold the name of the master channel
                        for i, grp in enumerate(self.groups):
                            group = f.create_group(r'/' + 'DataGroup_{}'.format(i + 1))

                            master_index = self.masters_db[i]

                            for j, ch in enumerate(grp['channels']):
                                name = ch.name
                                sig = self.get(group=i, index=j)
                                if j == master_index:
                                    group.attrs['master'] = name
                                dataset = group.create_dataset(name, data=sig.samples)
                                dataset.attrs['unit'] = sig.unit
                                dataset.attrs['comment'] = sig.comment if sig.comment else ''

            elif format == 'excel':
                try:
                    import xlsxwriter
                except ImportError:
                    warn('xlsxwriter not found; export to Excel is unavailable')
                    return
                else:
                    excel_name = os.path.splitext(name)[0]
                    nr = len(self.groups)
                    for i, grp in enumerate(self.groups):
                        print('Exporting group {} of {}'.format(i+1, nr))

                        workbook = xlsxwriter.Workbook('{}_{}.xlsx'.format(excel_name, 'DataGroup_{}'.format(i + 1)))
                        bold = workbook.add_format({'bold': True})

                        ws = workbook.add_worksheet("Information")

                        if self.version in MDF3_VERSIONS:
                            for j, item in enumerate(('date', 'time', 'author', 'organization', 'project', 'subject')):
                                ws.write(j, 0, item.title(), bold)
                                ws.write(j, 1, self.header[item].decode('latin-1'))

                            ws = workbook.add_worksheet('DataGroup_{}'.format(i + 1))

                            # the sheet header has 3 rows
                            # the channel name and unit 'YY [xx]'
                            # the channel comment
                            # the flag for data grup master channel
                            ws.write(0, 0, 'Channel', bold)
                            ws.write(1, 0, 'comment', bold)
                            ws.write(2, 0, 'is master', bold)

                            master_index = self.masters_db[i]

                            for j in range(grp['channel_group']['cycles_nr']):
                                ws.write(j+3, 0, str(j))

                            for j, ch in enumerate(grp['channels']):
                                sig = self.get(group=i, index=j)

                                col = j + 1
                                ws.write(0, col, '{} [{}]'.format(sig.name, sig.unit))
                                ws.write(1, col, sig.comment if sig.comment else '')
                                if j == master_index:
                                    ws.write(2, col, 'x')
                                ws.write_column(3, col, sig.samples.astype(str))

                        workbook.close()

            elif format == 'csv':
                csv_name = os.path.splitext(name)[0]
                nr = len(self.groups)
                for i, grp in enumerate(self.groups):
                    print('Exporting group {} of {}'.format(i+1, nr))
                    with open('{}_{}.csv'.format(csv_name, 'DataGroup_{}'.format(i + 1)), 'w', newline='') as csvfile:
                        writer = csv.writer(csvfile, delimiter=';')

                        ch_nr = len(grp['channels'])
                        channels = [self.get(group=i, index=j) for j in range(ch_nr)]

                        master_index = self.masters_db[i]
                        cycles = grp['channel_group']['cycles_nr']

                        writer.writerow(['Channel',] + ['{} [{}]'.format(ch.name, ch.unit) for ch in channels])
                        writer.writerow(['comment',] + [ch.comment for ch in channels])
                        writer.writerow(['is master',] + ['x' if j == master_index else '' for j in range(ch_nr)])

                        vals = [np.array(range(cycles), dtype=np.uint32), ] + [ch.samples for ch in channels]

                        writer.writerows(zip(*vals))

    def filter(self, channels):
        """ return new *MDF* object that contains only the channels listed in *channels* argument

        Parameters
        ----------
        channels : list
            list of channel names to be filtered

        Returns
        -------
        mdf : MDF
            new MDF file

        """

        # group channels by group index
        gps = {}
        for ch in channels:
            if ch in self.channels_db:
                for group, index in self.channels_db[ch]:
                    if not group in gps:
                        gps[group] = []
                    gps[group].append(index)
            else:
                warn('MDF filter error: Channel "{}" not found, it will be ignored'.format(ch))
                continue

        mdf = MDF(version=self.version)

        # append filtered channels to new MDF
        for group in gps:
            sigs = []
            for index in gps[group]:
                sigs.append(self.get(group=group, index=index))
            mdf.append(sigs,
                       'Signals filtered from <{}>'.format(os.path.basename(self.name) if self.name else 'New MDF'),
                       common_timebase=True)

        return mdf

    def iter_to_pandas(self):
        """ generator that yields channel groups as pandas DataFrames"""
        try:
            from pandas import DataFrame
        except ImportError:
            warn('pandas not found; export to pandas DataFrame is unavailable')
            return
        else:
            for i, gp in enumerate(self.groups):
                master_index = self.masters_db[i]
                pandas_dict = {gp['channels'][master_index].name: self.get(group=i, index=master_index, samples_only=True)}
                for j, ch in enumerate(gp['channels']):
                    if j == master_index:
                        continue
                    name = gp['channels'][j].name
                    pandas_dict[name] = self.get(group=i, index=j, samples_only=True)
                yield DataFrame.from_dict(pandas_dict)


if __name__ == '__main__':
    pass



"""
MDF v3 constants
"""
# byte order
BYTE_ORDER_INTEL = 0
BYTE_ORDER_MOTOROLA = 1

# data types
DATA_TYPE_UNSIGNED = 0
DATA_TYPE_SIGNED = 1
DATA_TYPE_FLOAT = 2
DATA_TYPE_DOUBLE = 3
DATA_TYPE_STRING = 7
DATA_TYPE_BYTEARRAY = 8
DATA_TYPE_UNSIGNED_INTEL = 13
DATA_TYPE_UNSIGNED_MOTOROLA = 9
DATA_TYPE_SIGNED_INTEL = 14
DATA_TYPE_SIGNED_MOTOROLA = 10
DATA_TYPE_FLOAT_INTEL = 15
DATA_TYPE_FLOAT_MOTOROLA = 11
DATA_TYPE_DOUBLE_INTEL = 16
DATA_TYPE_DOUBLE_MOTOROLA = 12

# channel types
CHANNEL_TYPE_VALUE = 0
CHANNEL_TYPE_MASTER = 1

# channel conversion types
CONVERSION_TYPE_NONE = 65535
CONVERSION_TYPE_LINEAR = 0
CONVERSION_TYPE_TABI = 1
CONVERSION_TYPE_TABX = 2
CONVERSION_TYPE_POLY = 6
CONVERSION_TYPE_EXPO = 7
CONVERSION_TYPE_LOGH = 8
CONVERSION_TYPE_RAT = 9
CONVERSION_TYPE_FORMULA = 10
CONVERSION_TYPE_VTAB = 11
CONVERSION_TYPE_VTABR = 12

# flags
FLAG_PRECISION = 1
FLAG_PHY_RANGE_OK = 2
FLAG_VAL_RANGE_OK = 8

# channel source types
SOURCE_ECU = 2
SOURCE_VECTOR = 19

# bus types
BUS_TYPE_NONE = 0
BUS_TYPE_CAN = 2
BUS_TYPE_FLEXRAY = 5

# file IO seek types
SEEK_START = 0
SEEK_REL = 1
SEEK_END = 2

# blocks size
ID_BLOCK_SIZE = 64
HEADER_COMMON_SIZE = 164
HEADER_POST_320_EXTRA_SIZE = 44
CE_BLOCK_SIZE = 128
FH_BLOCK_SIZE = 56
DG31_BLOCK_SIZE = 24
DG32_BLOCK_SIZE = 28
HD_BLOCK_SIZE = 104
CN_BLOCK_SIZE = 228
CG_BLOCK_SIZE = 26
CG33_BLOCK_SIZE = 30
DT_BLOCK_SIZE = 24
CC_COMMON_BLOCK_SIZE = 46
CC_ALG_BLOCK_SIZE = 88
CC_LIN_BLOCK_SIZE = 62
CC_POLY_BLOCK_SIZE = 94
CC_EXPO_BLOCK_SIZE = 102
CC_FORMULA_BLOCK_SIZE = 304
SR_BLOCK_SIZE = 156

# max int values
MAX_UINT8 = 2 << 8 - 1
MAX_UINT16 = 2 << 16 - 1
MAX_UNIT32 = 2 << 32 - 1
MAX_UINT64 = 2 << 64 - 1

# data location
LOCATION_ORIGINAL_FILE = 0
LOCATION_TEMPORARY_FILE = 1
LOCATION_MEMORY = 2

# blocks struct fmts and keys
ID_FMT = '<8s8s8s4H2s26s2H'
ID_KEYS = ('file_identification',
           'version_str',
           'program_identification',
           'byte_order',
           'float_format',
           'mdf_version',
           'code_page',
           'reserved0',
           'reserved1',
           'unfinalized_standard_flags',
           'unfinalized_custom_flags')

HEADER_COMMON_FMT = '<2sH3IH10s8s32s32s32s32s'
HEADER_COMMON_KEYS = ('id',
                      'block_len',
                      'first_dg_addr',
                      'comment_addr',
                      'program_addr',
                      'dg_nr',
                      'date',
                      'time',
                      'author',
                      'organization',
                      'project',
                      'subject')

HEADER_POST_320_EXTRA_FMT = 'Q2H32s'
HEADER_POST_320_EXTRA_KEYS = ('abs_time',
                              'tz_offset',
                              'time_quality',
                              'timer_identification')

FMT_CHANNEL = '<2sH5IH32s128s4H3d2IH'
KEYS_CHANNEL = ('id',
                'block_len',
                'next_ch_addr',
                'conversion_addr',
                'source_depend_addr',
                'ch_depend_addr',
                'comment_addr',
                'channel_type',
                'short_name',
                'description',
                'start_offset',
                'bit_count',
                'data_type',
                'range_flag',
                'min_raw_value',
                'max_raw_value',
                'sampling_rate',
                'long_name_addr',
                'display_name_addr',
                'aditional_byte_offset')

FMT_CHANNEL_GROUP = '<2sH3I3HI'
KEYS_CHANNEL_GROUP = ('id',
                      'block_len',
                      'next_cg_addr',
                      'first_ch_addr',
                      'comment_addr',
                      'record_id',
                      'ch_nr',
                      'samples_byte_nr',
                      'cycles_nr')

FMT_DATA_GROUP_32 = '<2sH4I2H4s'
KEYS_DATA_GROUP_32 = ('id',
                      'block_len',
                      'next_dg_addr',
                      'first_cg_addr',
                      'trigger_addr',
                      'data_block_addr',
                      'cg_nr',
                      'record_id_nr',
                      'reserved0')

FMT_DATA_GROUP = '<2sH4I2H'
KEYS_DATA_GROUP = ('id',
                   'block_len',
                   'next_dg_addr',
                   'first_cg_addr',
                   'trigger_addr',
                   'data_block_addr',
                   'cg_nr',
                   'record_id_nr')

FMT_SOURCE_COMMON = '<2s2H'
FMT_SOURCE_ECU = '<2s3HI80s32s4s'
FMT_SOURCE_EXTRA_ECU = '<HI80s32s4s'
KEYS_SOURCE_ECU = ('id',
                   'block_len',
                   'type',
                   'module_nr',
                   'module_address',
                   'description',
                   'ECU_identification',
                   'reserved0')

FMT_SOURCE_VECTOR = '<2s2H2I36s36s42s'
FMT_SOURCE_EXTRA_VECTOR = '<2I36s36s42s'
KEYS_SOURCE_VECTOR = ('id',
                      'block_len',
                      'type',
                      'CAN_id',
                      'CAN_ch_index',
                      'message_name',
                      'sender_name',
                      'reserved0')

KEYS_TEXT_BLOCK = ('id', 'block_len', 'text')

FMT_CONVERSION_COMMON = FMT_CONVERSION_NONE = '<2s2H2d20s2H'
FMT_CONVERSION_COMMON_SHORT = '<H2d20s2H'
KEYS_CONVESION_NONE = ('id',
                       'block_len',
                       'range_flag',
                       'min_phy_value',
                       'max_phy_value',
                       'unit',
                       'conversion_type',
                       'ref_param_nr')

FMT_CONVERSION_FORMULA = '<2s2H2d20s2H256s'
KEYS_CONVESION_FORMULA = ('id',
                          'block_len',
                          'range_flag',
                          'min_phy_value',
                          'max_phy_value',
                          'unit',
                          'conversion_type',
                          'ref_param_nr',
                          'formula')

FMT_CONVERSION_LINEAR = '<2s2H2d20s2H2d'
KEYS_CONVESION_LINEAR = ('id',
                         'block_len',
                         'range_flag',
                         'min_phy_value',
                         'max_phy_value',
                         'unit',
                         'conversion_type',
                         'ref_param_nr',
                         'b',
                         'a')

FMT_CONVERSION_POLY_RAT = '<2s2H2d20s2H6d'
KEYS_CONVESION_POLY_RAT = ('id',
                           'block_len',
                           'range_flag',
                           'min_phy_value',
                           'max_phy_value',
                           'unit',
                           'conversion_type',
                           'ref_param_nr',
                           'P1',
                           'P2',
                           'P3',
                           'P4',
                           'P5',
                           'P6')

FMT_CONVERSION_EXPO_LOGH = '<2s2H2d20s2H7d'
KEYS_CONVESION_EXPO_LOGH = ('id',
                            'block_len',
                            'range_flag',
                            'min_phy_value',
                            'max_phy_value',
                            'unit',
                            'conversion_type',
                            'ref_param_nr',
                            'P1',
                            'P2',
                            'P3',
                            'P4',
                            'P5',
                            'P6',
                            'P7')

FMT_PROGRAM_BLOCK = '<2sH{}s'
KEYS_PROGRAM_BLOCK = ('id', 'block_len', 'data')

FMT_SAMPLE_REDUCTION_BLOCK = '<2sH3Id'
KEYS_SAMPLE_REDUCTION_BLOCK = ('id',
                               'block_len',
                               'next_sr_addr',
                               'data_block_addr',
                               'cycles_nr',
                               'time_interval')



# -*- coding: utf-8 -*-
"""
asammdf *Signal* class module for time correct signal processing
"""
import numpy as np
import matplotlib.pyplot as plt

from .utils import MdfException


class Signal(object):
    """
    The Signal represents a signal described by it's samples and timestamps.
    It can do aritmethic operations agains other Signal or numeric type.
    The operations are computed in respect to the timestamps (time correct).
    The integer signals are not interpolated, instead the last value relative to the current timestamp is used.
    *samples*, *timstamps* and *name* are mandatory arguments.

    Parameters
    ----------
    samples : numpy.array | list | tuple
        signal samples
    timestamps : numpy.array | list | tuple
        signal timestamps
    unit : str
        signal unit
    name : str
        signal name
    conversion : dict
        dict describing the channel conversion , default *None*
    comment : str
        signal comment, default ''

    """
    def __init__(self, samples=None, timestamps=None, unit='', name='', conversion=None, comment=''):
        if samples is None or timestamps is None or name == '':
            raise MdfException('"samples", "timestamps" and "name" are mandatory arguments for Signal class instance')
        elif len(samples) != len(timestamps):
            raise MdfException('samples and timestamps lenght do not match ({} vs {})'.format(len(samples), len(timestamps)))
        else:
            if isinstance(samples, list) or isinstance(samples, tuple):
                samples = np.array(samples)
            if isinstance(timestamps, list) or isinstance(samples, tuple):
                timestamps = np.array(timestamps, dtype=np.float64)
            self.samples = samples
            self.timestamps = timestamps
            self.unit = unit
            self.name = name
            self.conversion = conversion
            self.comment = comment

    def __str__(self):
        return 'Signal {{ name="{}":\ts={}\tt={}\tunit="{}"\tconversion={} \tcomment="{}"}}'.format(self.name, self.samples, self.timestamps, self.unit, self.conversion, self.comment)

    def __repr__(self):
        return 'Signal {{ {}:\ts={}\tt={} }}'.format(self.name, repr(self.samples), repr(self.timestamps))

    def plot(self):
        """plot Signal samples"""
        fig = plt.figure()
        fig.canvas.set_window_title(self.name)
        if self.comment:
            plt.title('{}\n({})'.format(self.name, self.comment))
        else:
            plt.title(self.name)
        plt.xlabel('Time [s]')
        plt.ylabel('[{}]'.format(self.unit))
        plt.plot(self.timestamps, self.samples, 'b')
        plt.plot(self.timestamps, self.samples, 'b.')
        plt.grid(True)
        plt.show()

    def cut(self, start=None, stop=None):
        """
        Cuts the signal according to the *start* and *stop* values, by using the insertion indexes in the signal's *time* axis.

        Parameters
        ----------
        start : float
            start timestamp for cutting
        stop : float
            stop timestamp for cutting

        Returns
        -------
        result : Signal
            new *Signal* cut from the original

        Examples
        --------
        >>> new_sig = old_sig.cut(1.0, 10.5)
        >>> new_sig.timestamps[0], new_sig.timestamps[-1]
        0.98, 10.48

        """

        if start is None and stop is None:
            # return the channel uncut
            result = self

        else:
            if start is None:
                # cut from beggining to stop
                stop = np.searchsorted(self.timestamps, stop, side='right')
                if stop:
                    result = Signal(self.samples[: stop],
                                    self.timestamps[:stop],
                                    self.unit,
                                    self.name,
                                    self.conversion,
                                    self.comment)
                else:
                    result = Signal(np.array([]),
                                    np.array([]),
                                    self.unit,
                                    self.name,
                                    self.conversion,
                                    self.comment)

            elif stop is None:
                # cut from start to end
                start = np.searchsorted(self.timestamps, start, side='left')
                result = Signal(self.samples[start: ],
                                self.timestamps[start: ],
                                self.unit,
                                self.name,
                                self.conversion,
                                self.comment)

            else:
                # cut between start and stop
                start_ = np.searchsorted(self.timestamps, start, side='left')
                stop_ = np.searchsorted(self.timestamps, stop, side='right')
                if stop_ == start_:

                    if len(self.timestamps) and stop >= self.timestamps[0] and start <= self.timestamps[-1]:
                        # start and stop are found between 2 signal samples
                        # so return the previous sample
                        result = Signal(self.samples[start_ - 1: start_],
                                        self.timestamps[start_ - 1: start_],
                                        self.unit,
                                        self.name,
                                        self.conversion,
                                        self.comment)
                    else:
                        # signal is empty or start and stop are outside the signal time base
                        result = Signal(np.array([]),
                                        np.array([]),
                                        self.unit,
                                        self.name,
                                        self.conversion,
                                        self.comment)
                else:
                    result = Signal(self.samples[start_: stop_],
                                    self.timestamps[start_: stop_],
                                    self.unit,
                                    self.name,
                                    self.conversion,
                                    self.comment)
        return result

    def interp(self, new_timestamps):
        """ returns a new *Signal* interpolated using the *new_timestamps*"""
        if self.samples.dtype in ('float64', 'float32'):
            s = np.interp(new_timestamps, self.timestamps, self.samples)
        else:
            idx = np.searchsorted(self.timestamps, new_timestamps, side='right') - 1
            idx = np.clip(idx, 0, idx[-1])
            s = self.samples[idx]
        return Signal(s, new_timestamps, self.unit, self.name, self.conversion)

    def __apply_func(self, other, func_name):

        if isinstance(other, Signal):
            time = np.union1d(self.timestamps, other.timestamps)
            s = self.interp(time).samples
            o = other.interp(time).samples
            func = getattr(s, func_name)
            s = func(o)
        elif other is None:
            s = self.samples
            time = self.timestamps
        else:
            func = getattr(self.samples, func_name)
            s = func(other)
            time = self.timestamps
        return Signal(s, time, self.unit, self.name, self.conversion)

    def __pos__(self):
        return Signal(self.samples, self.timestamps, self.unit, self.name, self.conversion)

    def __neg__(self):
        return Signal(np.negative(self.samples), self.timestamps, self.unit, self.name, self.conversion)

    def __round__(self, n):
        return Signal(np.around(self.samples, n), self.timestamps, self.unit, self.name, self.conversion)

    def __sub__(self, other):
        return self.__apply_func(other, '__sub__')

    def __isub__(self, other):
        return self.__sub__(other)

    def __rsub__(self, other):
        return -self.__sub__(other)

    def __add__(self, other):
        return self.__apply_func(other, '__add__')

    def __iadd__(self, other):
        return self.__add__(other)

    def __radd__(self, other):
        return self.__add__(other)

    def __mul__(self, other):
        return self.__apply_func(other, '__mul__')

    def __imul__(self, other):
        return self.__mul__(other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return self.__apply_func(other, '__truediv__')

    def __itruediv__(self, other):
        return self.__truediv__(other)

    def __rtruediv__(self, other):
        return self.__apply_func(other, '__rtruediv__')

    def __mod__(self, other):
        return self.__apply_func(other, '__mod__')

    def __pow__(self, other):
        return self.__apply_func(other, '__pow__')

    def __and__(self, other):
        return self.__apply_func(other, '__and__')

    def __or__(self, other):
        return self.__apply_func(other, '__or__')

    def __xor__(self, other):
        return self.__apply_func(other, '__xor__')

    def __invert__(self):
        s = ~self.samples
        time = self.timestamps
        return Signal(s, time, self.unit, self.name, self.conversion)

    def __lshift__(self, other):
        return self.__apply_func(other, '__lshift__')

    def __rshift__(self, other):
        return self.__apply_func(other, '__rshift__')

    def __lt__(self, other):
        return self.__apply_func(other, '__lt__')

    def __le__(self, other):
        return self.__apply_func(other, '__le__')

    def __gt__(self, other):
        return self.__apply_func(other, '__gt__')

    def __ge__(self, other):
        return self.__apply_func(other, '__ge__')

    def __eq__(self, other):
        return self.__apply_func(other, '__eq__')

    def __ne__(self, other):
        return self.__apply_func(other, '__ne__')

    def __iter__(self):
        return zip(self.samples, self.timestamps)

    def __reversed__(self):
        return enumerate(zip(reversed(self.samples), reversed(self.timestamps)))

    def __len__(self):
        return len(self.samples)

    def __abs__(self):
        return Signal(np.fabs(self.samples), self.timestamps, self.unit, self.name, self.conversion)

    def __getitem__(self, val):
        return self.samples[val]

    def __setitem__(self, idx, val):
        self.samples[idx] = val

    def astype(self, np_type):
        """ returns new *Signal* with samples of dtype *np_type*"""
        return Signal(self.samples.astype(np_type), self.timestamps, self.unit, self.name, self.conversion)


if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
"""
MDF v4 constants
"""

DATA_TYPE_UNSIGNED_INTEL = 0
DATA_TYPE_UNSIGNED_MOTOROLA = 1
DATA_TYPE_SIGNED_INTEL = 2
DATA_TYPE_SIGNED_MOTOROLA = 3
DATA_TYPE_REAL_INTEL = 4
DATA_TYPE_REAL_MOTOROLA = 5
DATA_TYPE_STRING_LATIN_1 = 6
DATA_TYPE_STRING_UTF_8 = 7
DATA_TYPE_STRING_UTF_16_LE = 8
DATA_TYPE_STRING_UTF_16_BE = 9
DATA_TYPE_BYTEARRAY = 10
DATA_TYPE_MIME_SAMPLE = 11
DATA_TYPE_MIME_STREAM = 12
DATA_TYPE_CANOPEN_DATE = 13
DATA_TYPE_CANOPEN_TIME = 14

CHANNEL_TYPE_VALUE = 0
CHANNEL_TYPE_VLSD = 1
CHANNEL_TYPE_MASTER = 2
CHANNEL_TYPE_VIRTUAL_MASTER = 3
CHANNEL_TYPE_SYNC = 4
CHANNEL_TYPE_MLSD = 5
CHANNEL_TYPE_VIRTUAL = 6

CONVERSION_TYPE_NON = 0
CONVERSION_TYPE_LIN = 1
CONVERSION_TYPE_RAT = 2
CONVERSION_TYPE_ALG = 3
CONVERSION_TYPE_TABI = 4
CONVERSION_TYPE_TAB = 5
CONVERSION_TYPE_RTAB = 6
CONVERSION_TYPE_TABX = 7
CONVERSION_TYPE_RTABX = 8
CONVERSION_TYPE_TTAB = 9
CONVERSION_TYPE_TRANS = 10

CA_TYPE_ARRAY = 0
CA_TYPE_SCALE_AXIS = 1
CA_TYPE_LOOKUP = 2
CA_STORAGE_TYPE_CN_TEMPLATE = 0

SOURCE_ECU = 1
SOURCE_BUS = 2
SOURCE_IO = 3
SOURCE_TOOL = 4

BUS_TYPE_NONE = 0
BUS_TYPE_CAN = 2
BUS_TYPE_FLEXRAY = 5

SEEK_START = 0
SEEK_REL = 1
SEEK_END = 2

TIME_CH_SIZE = 8
SI_BLOCK_SIZE = 56
FH_BLOCK_SIZE = 56
DG_BLOCK_SIZE = 64
HD_BLOCK_SIZE = 104
CN_BLOCK_SIZE = 160
CG_BLOCK_SIZE = 104
COMMON_SIZE = 24
CC_NONE_BLOCK_SIZE = 80
CC_ALG_BLOCK_SIZE = 88
CC_LIN_BLOCK_SIZE = 96
AT_COMMON_SIZE = 96
DZ_COMMON_SIZE = 48
CC_COMMON_BLOCK_SIZE = 80
HL_BLOCK_SIZE = 40

FLAG_ALL_SAMPLES_VALID = 1
FLAG_INVALIDATION_BIT_VALID = 2

FLAG_PRECISION = 1
FLAG_PHY_RANGE_OK = 2
FLAG_VAL_RANGE_OK = 8
FLAG_AT_EMBEDDED = 1
FLAG_AT_COMPRESSED_EMBEDDED = 2
FLAG_AT_MD5_VALID = 4
FLAG_DZ_DEFLATE = 0
FLAG_DZ_TRANPOSED_DEFLATE = 1

# data location
LOCATION_ORIGINAL_FILE = 0
LOCATION_TEMPORARY_FILE = 1
LOCATION_MEMORY = 2

FMT_CHANNEL = '<4sI10Q4B4I2BH6d'
KEYS_CHANNEL = ('id',
                'reserved0',
                'block_len',
                'links_nr',
                'next_ch_addr',
                'component_addr',
                'name_addr',
                'source_addr',
                'conversion_addr',
                'data_block_addr',
                'unit_addr',
                'comment_addr',
                'channel_type',
                'sync_type',
                'data_type',
                'bit_offset',
                'byte_offset',
                'bit_count',
                'flags',
                'pos_invalidation_bit',
                'precision',
                'reserved1',
                'attachment_nr',
                'min_raw_value',
                'max_raw_value',
                'lower_limit',
                'upper_limit',
                'lower_ext_limit',
                'upper_ext_limit')

FMT_TEXT_BLOCK = '<4sIQQ{}s'
KEYS_TEXT_BLOCK = ('id',
                   'reserved0',
                   'block_len',
                   'links_nr',
                   'text')

FMT_SOURCE_INFORMATION = '<4sI5Q3B5s'
KEYS_SOURCE_INFORMATION = ('id',
                           'reserved0',
                           'block_len',
                           'links_nr',
                           'name_addr',
                           'path_addr',
                           'comment_addr',
                           'source_type',
                           'bus_type',
                           'flags',
                           'reserved1')

FMT_CHANNEL_GROUP = '<4sI10Q2H3I'
KEYS_CHANNEL_GROUP = ('id',
                      'reserved0',
                      'block_len',
                      'links_nr',
                      'next_cg_addr',
                      'first_ch_addr',
                      'acq_name_addr',
                      'acq_source_addr',
                      'first_sample_reduction_addr',
                      'comment_addr',
                      'record_id',
                      'cycles_nr',
                      'flags',
                      'path_separator',
                      'reserved1',
                      'samples_byte_nr',
                      'invalidation_bytes_nr')

FMT_DATA_BLOCK = '<4sI2Q{}s'
KEYS_DATA_BLOCK = ('id',
                   'reserved0',
                   'block_len',
                   'links_nr',
                   'data')

FMT_COMMON = '<4sI2Q'

FMT_FILE_HISTORY = '<4sI5Q2HB3s'
KEYS_FILE_HISTORY = ('id',
                     'reserved0',
                     'block_len',
                     'links_nr',
                     'next_fh_addr',
                     'comment_addr',
                     'abs_time',
                     'tz_offset',
                     'daylight_save_time',
                     'time_flags',
                     'reserved1')

FMT_DATA_GROUP = '<4sI6QB7s'
KEYS_DATA_GROUP = ('id',
                   'reserved0',
                   'block_len',
                   'links_nr',
                   'next_dg_addr',
                   'first_cg_addr',
                   'data_block_addr',
                   'comment_addr',
                   'record_id_len',
                   'reserved1')

FMT_DATA_LIST = '<4sIQQ{}QB3sIQ'

FMT_CONVERSION_NONE = '<4sI6Q2B3H2d'
KEYS_CONVERSION_NONE = ('id',
                        'reserved0',
                        'block_len',
                        'links_nr',
                        'name_addr',
                        'unit_addr',
                        'comment_addr',
                        'inv_conv_addr',
                        'conversion_type',
                        'precision',
                        'flags',
                        'ref_param_nr',
                        'val_param_nr',
                        'min_phy_value',
                        'max_phy_value')

FMT_CONVERSION_LINEAR = FMT_CONVERSION_NONE + '2d'
KEYS_CONVERSION_LINEAR = KEYS_CONVERSION_NONE + ('b', 'a')

FMT_CONVERSION_ALGEBRAIC = FMT_CONVERSION_NONE + 'Q'
KEYS_CONVERSION_ALGEBRAIC = KEYS_CONVERSION_NONE + ('formula_addr',)

FMT_CONVERSION_RAT = FMT_CONVERSION_NONE + '6d'
KEYS_CONVERSION_RAT = KEYS_CONVERSION_NONE + ('P1', 'P2', 'P3', 'P4', 'P5', 'P6')

FMT_HEADER_BLOCK = '<4sI9Q2H4B2Q'
FMT_IDENTIFICATION_BLOCK = '<8s8s8s5H26s2H'
IDENTIFICATION_BLOCK_SIZE = 64
HEADER_BLOCK_SIZE = 104
KEYS_HEADER_BLOCK = ('id',
                     'reserved3',
                     'block_len',
                     'links_nr',
                     'first_dg_addr',
                     'file_history_addr',
                     'channel_tree_addr',
                     'first_attachment_addr',
                     'first_event_addr',
                     'comment_addr',
                     'abs_time',
                     'tz_offset',
                     'daylight_save_time',
                     'time_flags',
                     'time_quality',
                     'flags',
                     'reserved4',
                     'start_angle',
                     'start_distance')
KEYS_IDENTIFICATION_BLOCK = ('file_identification',
                             'version_str',
                             'program_identification',
                             'reserved0',
                             'reserved1',
                             'mdf_version',
                             'reserved2',
                             'check_block',
                             'fill',
                             'unfinalized_standard_flags',
                             'unfinalized_custom_flags')

FMT_AT_COMMON = '<4sI6Q2HI16s2Q'
KEYS_AT_BLOCK = ('id',
                 'reserved0',
                 'block_len',
                 'links_nr',
                 'next_at_addr',
                 'file_name_addr',
                 'mime_addr',
                 'comment_addr',
                 'flags',
                 'creator_index',
                 'reserved1',
                 'md5_sum',
                 'original_size',
                 'embedded_size',
                 'embedded_data')

FMT_DZ_COMMON = '<4sI2Q2s2BI2Q'
KEYS_DZ_BLOCK = ('id',
                 'reserved0',
                 'block_len',
                 'links_nr',
                 'original_type',
                 'zip_type',
                 'reserved1',
                 'param',
                 'original_size',
                 'zip_size',
                 'data')

FMT_HL_BLOCK = '<4sI3QHB5s'
KEYS_HL_BLOCK = ('id',
                 'reserved0',
                 'block_len',
                 'links_nr',
                 'first_dl_addr',
                 'flags',
                 'zip_type',
                 'reserved1')




"""
ASAM MDF version 3 file format module

"""
from __future__ import print_function, division
import sys
PYVERSION = sys.version_info[0]

import os
import time
import warnings

from collections import defaultdict
from functools import reduce
from tempfile import TemporaryFile

from numpy import (interp, linspace, dtype, amin, amax, array_equal,
                   array, searchsorted, log, exp, clip, union1d, float64,
                   uint8, frombuffer, issubdtype, flexible, arange, recarray)
from numpy.core.records import fromstring, fromarrays
from numexpr import evaluate

from .utils import MdfException, get_fmt, pair, fmt_to_datatype
from .signal import Signal
from .signalconstants import *
from .v3constants import *
from .v3blocks import (Channel, ChannelConversion, ChannelDependency,
                       ChannelExtension, ChannelGroup, DataBlock, DataGroup,
                       FileIdentificationBlock, HeaderBlock, ProgramBlock,
                       SampleReduction, TextBlock, TriggerBlock)


if PYVERSION == 2:
    def bytes(obj):
        return obj.__bytes__()


__all__ = ['MDF3', ]


class MDF3(object):
    """If the *name* exist it will be loaded otherwise an empty file will be created that can be later saved to disk

    Parameters
    ----------
    name : string
        mdf file name
    load_measured_data : bool
        load data option; default *True*

        * if *True* the data group binary data block will be loaded in RAM
        * if *False* the channel data is read from disk on request

    version : string
        mdf file version ('3.00', '3.10', '3.20' or '3.30'); default '3.20'

    Attributes
    ----------
    name : string
        mdf file name
    groups : list
        list of data groups
    header : OrderedDict
        mdf file header
    file_history : TextBlock
        file history text block; can be None
    load_measured_data : bool
        load measured data option
    version : str
        mdf version
    channels_db : dict
        used for fast channel access by name; for each name key the value is a list of (group index, channel index) tuples
    masters_db : dict
        used for fast master channel access; for each group index key the value is the master channel index

    """

    def __init__(self, name=None, load_measured_data=True, version='3.20'):
        self.groups = []
        self.header = None
        self.identification = None
        self.file_history = None
        self.name = name
        self.load_measured_data = load_measured_data
        self.channels_db = {}
        self.masters_db = {}

        # used when appending to MDF object created with load_measured_data=False
        self._tempfile = None

        if name:
            self._read()
        else:
            self.identification = FileIdentificationBlock(version=version)
            self.version = version
            self.header = HeaderBlock(version=self.version)

    def _load_group_data(self, group):
        """ get group's data block bytes"""
        if self.load_measured_data == False:
            # could be an appended group
            # for now appended groups keep the measured data in the memory.
            # the plan is to use a temp file for appended groups, to keep the
            # memory usage low.
            if group['data_location'] == LOCATION_ORIGINAL_FILE:
                # this is a group from the source file
                # so fetch the measured data from it
                with open(self.name, 'rb') as file_stream:
                    # go to the first data block of the current data group
                    dat_addr = group['data_group']['data_block_addr']

                    if group.get('sorted', True):
                        read_size = group['size']
                        data = DataBlock(file_stream=file_stream, address=dat_addr, size=read_size)['data']

                    else:
                        read_size = group['size']
                        record_id = group['channel_group']['record_id']
                        cg_size = group['record_size']
                        record_id_nr = group['data_group']['record_id_nr'] if group['data_group']['record_id_nr'] <= 2 else 0
                        cg_data = []

                        data = DataBlock(file_stream=file_stream, address=dat_addr, size=read_size)['data']

                        i = 0
                        size = len(data)
                        while i < size:
                            rec_id = data[i]
                            # skip redord id
                            i += 1
                            rec_size = cg_size[rec_id]
                            if rec_id == record_id:
                                rec_data = data[i: i+rec_size]
                                cg_data.append(rec_data)
                            # if 2 record id's are sued skip also the second one
                            if record_id_nr == 2:
                                i += 1
                            # go to next record
                            i += rec_size
                        data = b''.join(cg_data)
            elif group['data_location'] == LOCATION_TEMPORARY_FILE:
                read_size = group['size']
                dat_addr = group['data_group']['data_block_addr']
                self._tempfile.seek(dat_addr, SEEK_START)
                data = self._tempfile.read(read_size)
        else:
            data = group['data_block']['data']
        return data

    def _prepare_record(self, group):
        """ compute record dtype and parents dict for this group

        Parameters
        ----------
        group : dict
            MDF group dict

        Returns
        -------
        parents, dtypes : dict, numpy.dtype
            mapping of channels to records fields, records fiels dtype

        """
        grp = group
        record_size = grp['channel_group']['samples_byte_nr'] << 3
        next_byte_aligned_position = 0
        types = []
        current_parent = ""
        parent_start_offset = 0
        parents = {}
        group_channels = set()

        # the channels are first sorted ascending (see __lt__ method of Channel class):
        # a channel with lower start offset is smaller, when two channels have
        # the same start offset the one with higer bit size is considered smaller.
        # The reason is that when the numpy record is built and there are overlapping
        # channels, the parent fields should be bigger (bit size) than the embedded
        # channels. For each channel the parent dict will have a (parent name, bit offset) pair:
        # the channel value is computed using the values from the parent field,
        # and the bit offset, which is the channel's bit offset within the parent bytes.
        # This means all parents will have themselves as parent, and bit offset of 0.
        # Gaps in the records are also considered. Non standard integers size is
        # adjusted to the first higher standard integer size (eq. uint of 28bits will
        # be adjusted to 32bits)

        for original_index, new_ch in sorted(enumerate(grp['channels']), key=lambda i: i[1]):
            # channels with channel dependencies are skipped from the numpy record
            if new_ch['ch_depend_addr']:
                continue

            start_offset = new_ch['start_offset']
            bit_offset = start_offset % 8
            data_type = new_ch['data_type']
            bit_count = new_ch['bit_count']
            name = str(new_ch.name)

            # handle multiple occurance of same channel name
            i = 0
            new_name = name
            while new_name in group_channels:
                new_name = "{}_{}".format(name, i)
                i += 1
            name = str(new_name)
            group_channels.add(name)

            if start_offset >= next_byte_aligned_position:
                parent_start_offset = (start_offset >> 3 ) << 3
                parents[original_index] = name, bit_offset

                # check if there are byte gaps in the record
                gap = (parent_start_offset - next_byte_aligned_position) >> 3
                if gap:
                    types.append( ('', 'a{}'.format(gap)) )

                # adjust size to 1, 2, 4 or 8 bytes for nonstandard integers
                size = bit_offset + bit_count
                if data_type != DATA_TYPE_STRING:
                    if size > 32:
                        next_byte_aligned_position = parent_start_offset + 64
                        size = 8
                    elif size > 16:
                        next_byte_aligned_position = parent_start_offset + 32
                        size = 4
                    elif size > 8:
                        next_byte_aligned_position = parent_start_offset + 16
                        size = 2
                    else:
                        next_byte_aligned_position = parent_start_offset + 8
                        size = 1
                else:
                    size = size >> 3

                types.append( (name, get_fmt(data_type, size)) )

                current_parent = name
            else:
                parents[original_index] = current_parent, start_offset - parent_start_offset
        gap = (record_size - next_byte_aligned_position) >> 3
        if gap:
            types.append( ('', 'a{}'.format(gap)) )

        return parents, dtype(types)

    def _read(self):
        with open(self.name, 'rb') as file_stream:

            # performance optimization
            read = file_stream.read
            seek = file_stream.seek

            dg_cntr = 0
            seek(0, SEEK_START)

            self.identification = FileIdentificationBlock(file_stream=file_stream)
            self.header = HeaderBlock(file_stream=file_stream)

            self.version = self.identification['version_str'].decode('latin-1').strip(' \n\t\x00')

            self.file_history = TextBlock(address=self.header['comment_addr'], file_stream=file_stream)

            # this will hold mapping from channel address to Channel object
            # needed for linking dependecy blocks to refernced channels after the file is loaded
            ch_map = {}

            # go to first date group
            dg_addr = self.header['first_dg_addr']
            # read each data group sequentially
            while dg_addr:
                gp = DataGroup(address=dg_addr, file_stream=file_stream)
                cg_nr = gp['cg_nr']
                cg_addr = gp['first_cg_addr']
                data_addr = gp['data_block_addr']

                # read trigger information if available
                trigger_addr = gp['trigger_addr']
                if trigger_addr:
                    trigger = TriggerBlock(address=trigger_addr, file_stream=file_stream)
                    if trigger['comment_addr']:
                        trigger_text = TextBlock(address=trigger['comment_addr'], file_stream=file_stream)
                    else:
                        trigger_text = None
                else:
                    trigger = None
                    trigger_text = None

                new_groups = []
                for i in range(cg_nr):

                    new_groups.append({})
                    grp = new_groups[-1]
                    grp['channels'] = []
                    grp['channel_conversions'] = []
                    grp['channel_extensions'] = []
                    grp['data_block'] = None
                    grp['texts'] = {'channels': [], 'conversion_tab': [], 'channel_group': []}
                    grp['trigger'] = [trigger, trigger_text]
                    grp['channel_dependencies'] = []

                    kargs = {'first_cg_addr': cg_addr,
                             'data_block_addr': data_addr}
                    if self.version in ('3.20', '3.30'):
                        kargs['block_len'] = DG32_BLOCK_SIZE
                    else:
                        kargs['block_len'] = DG31_BLOCK_SIZE

                    grp['data_group'] = DataGroup(**kargs)

                    # read each channel group sequentially
                    grp['channel_group'] = ChannelGroup(address=cg_addr, file_stream=file_stream)

                    # read acquisition name and comment for current channel group
                    grp['texts']['channel_group'].append({})

                    address = grp['channel_group']['comment_addr']
                    if address:
                        grp['texts']['channel_group'][-1]['comment_addr'] = TextBlock(address=address, file_stream=file_stream)

                    # go to first channel of the current channel group
                    ch_addr = grp['channel_group']['first_ch_addr']
                    ch_cntr = 0
                    grp_chs = grp['channels']
                    grp_conv = grp['channel_conversions']
                    grp_ch_texts = grp['texts']['channels']

                    while ch_addr:
                        # read channel block and create channel object
                        new_ch = Channel(address=ch_addr, file_stream=file_stream)

                        # check if it has channel dependencies
                        if new_ch['ch_depend_addr']:
                            grp['channel_dependencies'].append(ChannelDependency(address=new_ch['ch_depend_addr'], file_stream=file_stream))
                        else:
                            grp['channel_dependencies'].append(None)

                        # update channel map
                        ch_map[ch_addr] = (new_ch, grp)

                        # read conversion block and create channel conversion object
                        address = new_ch['conversion_addr']
                        if address:
                            new_conv = ChannelConversion(address=address, file_stream=file_stream)
                            grp_conv.append(new_conv)
                        else:
                            new_conv = None
                            grp_conv.append(None)

                        vtab_texts = {}
                        if new_conv and new_conv['conversion_type'] == CONVERSION_TYPE_VTABR:
                            for idx in range(new_conv['ref_param_nr']):
                                address = new_conv['text_{}'.format(idx)]
                                if address:
                                    vtab_texts['text_{}'.format(idx)] = TextBlock(address=address, file_stream=file_stream)
                        grp['texts']['conversion_tab'].append(vtab_texts)

                        if self.load_measured_data:
                            # read source block and create source infromation object
                            address = new_ch['source_depend_addr']
                            if address:
                                grp['channel_extensions'].append(ChannelExtension(address=address, file_stream=file_stream))
                            else:
                                grp['channel_extensions'].append(None)
                        else:
                            grp['channel_extensions'].append(None)

                        # read text fields for channel
                        ch_texts = {}
                        for key in ('long_name_addr', 'comment_addr', 'display_name_addr'):
                            address = new_ch[key]
                            if address:
                                ch_texts[key] = TextBlock(address=address, file_stream=file_stream)
                        grp_ch_texts.append(ch_texts)

                        # update channel object name and block_size attributes
                        if new_ch['long_name_addr']:
                            new_ch.name = ch_texts['long_name_addr'].text_str
                        else:
                            new_ch.name = new_ch['short_name'].decode('latin-1').strip(' \n\t\x00')

                        if new_ch.name in self.channels_db:
                            self.channels_db[new_ch.name].append((dg_cntr, ch_cntr))
                        else:
                            self.channels_db[new_ch.name] = []
                            self.channels_db[new_ch.name].append((dg_cntr, ch_cntr))

                        if new_ch['channel_type'] == CHANNEL_TYPE_MASTER:
                            self.masters_db[dg_cntr] = ch_cntr
                        # go to next channel of the current channel group
                        ch_addr = new_ch['next_ch_addr']
                        ch_cntr += 1
                        grp_chs.append(new_ch)

                    cg_addr = grp['channel_group']['next_cg_addr']
                    dg_cntr += 1

                # store channel groups record sizes dict and data block size in each
                # new group data belong to the initial unsorted group, and add
                # the key 'sorted' with the value False to use a flag;
                # this is used later if load_measured_data=False

                if cg_nr > 1:
                    # this is an unsorted file since there are multiple channel groups
                    # within a data group
                    cg_size = {}
                    size = 0
                    record_id_nr = gp['record_id_nr'] if gp['record_id_nr'] <= 2 else 0
                    for grp in new_groups:
                        size += (grp['channel_group']['samples_byte_nr'] + record_id_nr) * grp['channel_group']['cycles_nr']
                        cg_size[grp['channel_group']['record_id']] = grp['channel_group']['samples_byte_nr']

                    for grp in new_groups:
                        grp['sorted'] = False
                        grp['record_size'] = cg_size
                        grp['size'] = size
                else:
                    record_id_nr = gp['record_id_nr'] if gp['record_id_nr'] <= 2 else 0
                    grp['size'] = size = (grp['channel_group']['samples_byte_nr'] + record_id_nr) * grp['channel_group']['cycles_nr']

                if self.load_measured_data:
                    grp['data_location'] = LOCATION_MEMORY
                    # read data block of the current data group
                    dat_addr = gp['data_block_addr']
                    if dat_addr:
                        seek(dat_addr, SEEK_START)
                        data = read(size)
                    else:
                        data = b''
                    if cg_nr == 1:
                        grp = new_groups[0]
                        kargs = {'data': data}
                        grp['data_block'] = DataBlock(**kargs)

                    else:
                        # agregate data for each record ID in the cg_data dict
                        cg_data = defaultdict(list)
                        i = 0
                        size = len(data)
                        while i < size:
                            rec_id = data[i]
                            # skip redord id
                            i += 1
                            rec_size = cg_size[rec_id]
                            rec_data = data[i: i+rec_size]
                            cg_data[rec_id].append(rec_data)
                            # if 2 record id's are sued skip also the second one
                            if record_id_nr == 2:
                                i += 1
                            # go to next record
                            i += rec_size
                        for grp in new_groups:
                            kargs = {}
                            kargs['data'] = b''.join(cg_data[grp['channel_group']['record_id']])
                            grp['channel_group']['record_id'] = 1
                            grp['data_block'] = DataBlock(**kargs)
                else:
                    grp['data_location'] = LOCATION_ORIGINAL_FILE

                self.groups.extend(new_groups)

                # go to next data group
                dg_addr = gp['next_dg_addr']

            # once the file has been loaded update the channel depency refenreces
            for grp in self.groups:
                for dependency_block in grp['channel_dependencies']:
                    if dependency_block:
                        for i in range(dependency_block['sd_nr']):
                            ref_channel_addr = dependency_block['ch_{}'.format(i)]
                            dependency_block.referenced_channels.append(ch_map[ref_channel_addr])

    def add_trigger(self, group, time, pre_time=0, post_time=0, comment=''):
        """ add trigger to data group

        Parameters
        ----------
        group : int
            group index
        time : float
            trigger time
        pre_time : float
            trigger pre time; default 0
        post_time : float
            trigger post time; default 0
        comment : str
            trigger comment

        """
        gp = self.groups[group]
        trigger, trigger_text = gp['trigger']
        if trigger:
            nr = trigger['trigger_event_nr']
            trigger['trigger_event_nr'] += 1
            trigger['block_len'] += 24
            trigger['trigger_{}_time'.format(nr)] = time
            trigger['trigger_{}_pretime'.format(nr)] = pre_time
            trigger['trigger_{}_posttime'.format(nr)] = post_time
            if trigger_text is None and comment:
                trigger_text = TextBlock.from_text(comment)
                gp['trigger'][1] = trigger_text
        else:
            trigger = TriggerBlock(trigger_event_nr=1,
                                   trigger_0_time=time,
                                   trigger_0_pretime=pre_time,
                                   trigger_0_posttime=post_time)
            if comment:
                trigger_text = TextBlock.from_text(comment)
            else:
                trigger_text = None

            gp['trigger'] = [trigger, trigger_text]

    def append(self, signals, acquisition_info='Python', common_timebase=False):
        """
        Appends a new data group.

        For channel depencies type Signals, the *samples* attribute must be a numpy.recarray

        Parameters
        ----------
        signals : list
            list on *Signal* objects
        acquisition_info : str
            acquisition information; default 'Python'
        common_timebase : bool
            flag to hint that the signals have the same timebase

        Examples
        --------
        >>> # case 1 conversion type None
        >>> s1 = np.array([1, 2, 3, 4, 5])
        >>> s2 = np.array([-1, -2, -3, -4, -5])
        >>> s3 = np.array([0.1, 0.04, 0.09, 0.16, 0.25])
        >>> t = np.array([0.001, 0.002, 0.003, 0.004, 0.005])
        >>> names = ['Positive', 'Negative', 'Float']
        >>> units = ['+', '-', '.f']
        >>> info = {}
        >>> s1 = Signal(samples=s1, timstamps=t, unit='+', name='Positive')
        >>> s2 = Signal(samples=s2, timstamps=t, unit='-', name='Negative')
        >>> s3 = Signal(samples=s3, timstamps=t, unit='flts', name='Floats')
        >>> mdf = MDF3('new.mdf')
        >>> mdf.append([s1, s2, s3], 'created by asammdf v1.1.0')
        >>> # case 2: VTAB conversions from channels inside another file
        >>> mdf1 = MDF3('in.mdf')
        >>> ch1 = mdf1.get("Channel1_VTAB")
        >>> ch2 = mdf1.get("Channel2_VTABR")
        >>> sigs = [ch1, ch2]
        >>> mdf2 = MDF3('out.mdf')
        >>> mdf2.append(sigs, 'created by asammdf v1.1.0')

        """
        dg_cntr = len(self.groups)

        gp = {}
        gp['channels'] = gp_channels = []
        gp['channel_conversions'] = gp_conv = []
        gp['channel_extensions'] = gp_source = []
        gp['channel_dependencies'] = gp_dep = []
        gp['texts'] = gp_texts = {'channels': [],
                                  'conversion_tab': [],
                                  'channel_group': []}
        self.groups.append(gp)

        if not signals:
            raise MdfException('"append" requires a non-empty list of Signal objects')

        # check if the signals have a common timebase
        # if not interpolate the signals using the union of all timbases
        t_ = signals[0].timestamps
        if not common_timebase:
            for s in signals[1:]:
                if not array_equal(s.timestamps, t_):
                    different = True
                    break
            else:
                different = False

            if different:
                times = [s.timestamps for s in signals]
                t = reduce(union1d, times).flatten().astype(float64)
                signals = [s.interp(t) for s in signals]
                times = None
            else:
                t = t_
        else:
            t = t_

        # split for signals that come from a channel dependency
        # (this means the samples will be of type np.recarray)
        # from the regular one dimensional signals.
        # The regular signals will be first added to the group.
        # The recarray signals will be saved along side the fields, which will
        # be saved as new signals.
        simple_signals = []
        recarray_signals = []
        channel_nr = 0
        for sig in signals:
            if isinstance(sig.samples, recarray):
                recarray_signals.append(sig)
                channel_nr += len(sig.samples.dtype) + 1
            else:
                simple_signals.append(sig)
                channel_nr += 1

        cycles_nr = len(t)

        # setup all blocks related to the time master channel
        t_type, t_size = fmt_to_datatype(t.dtype)

        #time channel texts
        for _, item in gp_texts.items():
            item.append({})

        gp_texts['channel_group'][-1]['comment_addr'] = TextBlock.from_text(acquisition_info)

        #conversion for time channel
        kargs = {'conversion_type': CONVERSION_TYPE_NONE,
                 'unit': 's'.encode('latin-1'),
                 'min_phy_value': t[0] if cycles_nr else 0,
                 'max_phy_value': t[-1] if cycles_nr else 0}
        gp_conv.append(ChannelConversion(**kargs))

        #source for time
        kargs = {'module_nr': 0,
                 'module_address': 0,
                 'type': SOURCE_ECU,
                 'description': 'Channel inserted by Python Script'.encode('latin-1')}
        gp_source.append(ChannelExtension(**kargs))

        #time channel
        kargs = {'short_name': 't'.encode('latin-1'),
                 'channel_type': CHANNEL_TYPE_MASTER,
                 'data_type': t_type,
                 'start_offset': 0,
                 'min_raw_value' : t[0] if cycles_nr else 0,
                 'max_raw_value' : t[-1] if cycles_nr else 0,
                 'bit_count': t_size}
        ch = Channel(**kargs)
        ch.name = 't'
        gp_channels.append(ch)
        self.masters_db[dg_cntr] = 0

        # time channel doesn't have channel dependencies
        gp_dep.append(None)

        # prepare start bit offset and channel counter for the other channels
        offset = t_size
        ch_cntr = 1

        # arrays will hold the samples for all channels
        # types holds the (channel name, numpy dtype) pairs for all channels
        # formats holds the (ASAM channel data type, bit size) for all channels
        arrays = [t, ]
        types = [('t', t.dtype),]
        formats = [fmt_to_datatype(t.dtype), ]

        # data group record parents
        parents = {0: ('t', 0)}

        # first add the signals in the simple signal list
        if simple_signals:
            #channels texts
            names = [s.name for s in simple_signals]
            for name in names:
                for _, item in gp['texts'].items():
                    item.append({})
                if len(name) >= 32:
                    gp_texts['channels'][-1]['long_name_addr'] = TextBlock.from_text(name)

            # conversions for channels
            if cycles_nr:
                min_max = []
                # compute min and max values for all channels
                # for string channels we append (1,0) and use this as a marker (if min>max then channel is string)
                for s in simple_signals:
                    if issubdtype(s.samples.dtype, flexible):
                        min_max.append((1,0))
                    else:
                        min_max.append((amin(s.samples), amax(s.samples)))
            else:
                min_max = [(0, 0) for s in simple_signals]

            #conversion for channels
            for idx, s in enumerate(simple_signals):
                conv = s.conversion
                if conv:
                    conv_type = conv['type']
                    if conv_type in (SIGNAL_TYPE_V3_VTAB, SIGNAL_TYPE_V4_VTAB):
                        kargs = {}
                        kargs['conversion_type'] = CONVERSION_TYPE_VTAB
                        raw = conv['raw']
                        phys = conv['phys']
                        for i, (r_, p_) in enumerate(zip(raw, phys)):
                            kargs['text_{}'.format(i)] = p_[:31] + b'\x00'
                            kargs['param_val_{}'.format(i)] = r_
                        kargs['ref_param_nr'] = len(raw)
                        kargs['unit'] = s.unit.encode('latin-1')
                    elif conv_type in (SIGNAL_TYPE_V3_VTABR, SIGNAL_TYPE_V4_VTABR):
                        kargs = {}
                        kargs['conversion_type'] = CONVERSION_TYPE_VTABR
                        lower = conv['lower']
                        upper = conv['upper']
                        texts = conv['phys']
                        kargs['unit'] = s.unit.encode('latin-1')
                        kargs['ref_param_nr'] = len(upper)

                        for i, (u_, l_, t_) in enumerate(zip(upper, lower, texts)):
                            kargs['lower_{}'.format(i)] = l_
                            kargs['upper_{}'.format(i)] = u_
                            kargs['text_{}'.format(i)] = 0
                            gp_texts['conversion_tab'][-1]['text_{}'.format(i)] = TextBlock.from_text(t_)

                    else:
                         kargs = {'conversion_type': CONVERSION_TYPE_NONE,
                                  'unit': s.unit.encode('latin-1'),
                                  'min_phy_value': min_max[idx][0],
                                  'max_phy_value': min_max[idx][1]}
                    gp_conv.append(ChannelConversion(**kargs))
                else:
                    kargs = {'conversion_type': CONVERSION_TYPE_NONE,
                             'unit': s.unit.encode('latin-1') ,
                             'min_phy_value': min_max[idx][0] if min_max[idx][0]<=min_max[idx][1] else 0,
                             'max_phy_value': min_max[idx][1] if min_max[idx][0]<=min_max[idx][1] else 0}
                    gp_conv.append(ChannelConversion(**kargs))

            #source for channels
            for _ in simple_signals:
                kargs = {'module_nr': 0,
                         'module_address': 0,
                         'type': SOURCE_ECU,
                         'description': 'Channel inserted by Python Script'.encode('latin-1')}
                gp_source.append(ChannelExtension(**kargs))

            sig_dtypes = [s.samples.dtype for s in simple_signals]
            sig_formats = [fmt_to_datatype(typ) for typ in sig_dtypes]

            #channels
            for (sigmin, sigmax), (sig_type, sig_size), s in zip(min_max, sig_formats, simple_signals):
                # compute additional byte offset for large records size
                if offset > MAX_UINT16:
                    additional_byte_offset = (offset - MAX_UINT16 ) >> 3
                    start_bit_offset = offset - additional_byte_offset << 3
                else:
                    start_bit_offset = offset
                    additional_byte_offset = 0
                kargs = {'short_name': (s.name[:31] + '\x00').encode('latin-1') if len(s.name) >= 32 else s.name.encode('latin-1'),
                         'channel_type': CHANNEL_TYPE_VALUE,
                         'data_type': sig_type,
                         'min_raw_value': sigmin if sigmin <= sigmax else 0,
                         'max_raw_value': sigmax if sigmin <= sigmax else 0,
                         'start_offset': start_bit_offset,
                         'bit_count': sig_size,
                         'aditional_byte_offset' : additional_byte_offset}
                if s.comment:
                    kargs['description'] = (s.comment[:127] + '\x00').encode('latin-1') if len(s.comment) >= 128 else s.comment.encode('latin-1')
                ch = Channel(**kargs)

                name = s.name

                ch.name = name
                gp_channels.append(ch)
                offset += sig_size

                if not name in self.channels_db:
                    self.channels_db[name] = []
                self.channels_db[name].append((dg_cntr, ch_cntr))

                # update the parents as well
                parents[ch_cntr] = name, 0

                ch_cntr += 1


            # simple channels don't have channel dependencies
            for _ in simple_signals:
                gp_dep.append(None)

            # extend arrays, types and formats with data related to the simple signals
            arrays.extend(s.samples for s in simple_signals)
            types.extend([(name, typ) for name, typ in zip(names, sig_dtypes)])
            formats.extend(sig_formats)

        # second, add the recarray signals
        if recarray_signals:
            for sig in recarray_signals:
                names = sig.samples.dtype.names
                signals = [sig.samples[name] for name in names]

                # numpy dtype does not allow for reapeating names so we must handle name conflincts
                dtype_fields = [t[0] for t in types]
                new_names = []
                # if the name already exist in the dtype fiels then
                # compute a new name "name_xx", by incrementing the index
                # until a valid new name is found
                for name in names:
                    i = 0
                    new_name = name
                    while new_name in dtype_fields:
                        new_name = "{}_{}".format(name, i)
                        i += 1
                    new_names.append(new_name)

                names = new_names

                # add recarray parent signal texts
                name = sig.name
                for _, item in gp['texts'].items():
                    item.append({})
                if len(name) >= 32:
                    gp_texts['channels'][-1]['long_name_addr'] = TextBlock.from_text(name)
                # add components texts
                for name in names:
                    for _, item in gp['texts'].items():
                        item.append({})
                    if len(name) >= 32:
                        gp_texts['channels'][-1]['long_name_addr'] = TextBlock.from_text(name)

                # recarray parent has no conversion
                gp_conv.append(None)
                # add components conversions
                min_max = []
                if cycles_nr:
                    for s in signals:
                        min_max.append( (amin(s), amax(s)) )
                else:
                    for s in signals:
                        min_max = [(0, 0)]
                for i, s in enumerate(signals):
                    kargs = {'conversion_type': CONVERSION_TYPE_NONE,
                             'unit': b'',
                             'min_phy_value': min_max[i][0],
                             'max_phy_value': min_max[i][1]}
                    gp_conv.append(ChannelConversion(**kargs))

                # add parent and components sources
                kargs = {'module_nr': 0,
                         'module_address': 0,
                         'type': SOURCE_ECU,
                         'description': 'Channel inserted by Python Script'.encode('latin-1')}
                gp_source.append(ChannelExtension(**kargs))
                for _ in signals:
                    gp_source.append(ChannelExtension(**kargs))

                new_sig_dtypes = [s.dtype for s in signals]
                new_sig_formats = [fmt_to_datatype(typ) for typ in new_sig_dtypes]

                # extend arrays, types and formasts with data from current recarray
                arrays.extend(signals)
                types.extend([(name, dtype_) for name, dtype_ in zip(names, new_sig_dtypes)])
                formats.extend(new_sig_formats)

                # add channel dependency block for recarray parent channel
                parent_dep = ChannelDependency(sd_nr=len(signals))
                gp_dep.append(parent_dep)

                # components do not have channel dependencies
                for _ in signals:
                    gp_dep.append(None)

                # add parent channel
                if offset > MAX_UINT16:
                    additional_byte_offset = (offset - MAX_UINT16 ) >> 3
                    start_bit_offset = offset - additional_byte_offset << 3
                else:
                    start_bit_offset = offset
                    additional_byte_offset = 0
                kargs = {'short_name': (sig.name[:31] + '\x00').encode('latin-1') if len(sig.name) >= 32 else sig.name.encode('latin-1'),
                         'channel_type': CHANNEL_TYPE_VALUE,
                         'data_type': new_sig_formats[0][0],
                         'min_raw_value': 0,
                         'max_raw_value': 0,
                         'start_offset': start_bit_offset,
                         'bit_count': new_sig_formats[0][1],
                         'aditional_byte_offset' : additional_byte_offset}
                ch = Channel(**kargs)
                ch.name = name = sig.name
                gp_channels.append(ch)

                # update channel database with racaaray parent indexes
                if not name in self.channels_db:
                    self.channels_db[name] = []
                self.channels_db[name].append((dg_cntr, ch_cntr))

                ch_cntr += 1

                # add components channels
                for (sigmin, sigmax), (sig_type, sig_size), name in zip(min_max, new_sig_formats, names):
                    if offset > MAX_UINT16:
                        additional_byte_offset = (offset - MAX_UINT16 ) >> 3
                        start_bit_offset = offset - additional_byte_offset << 3
                    else:
                        start_bit_offset = offset
                        additional_byte_offset = 0
                    kargs = {'short_name': (name[:31] + '\x00').encode('latin-1') if len(name) >= 32 else name.encode('latin-1'),
                             'channel_type': CHANNEL_TYPE_VALUE,
                             'data_type': sig_type,
                             'min_raw_value': sigmin if sigmin<=sigmax else 0,
                             'max_raw_value': sigmax if sigmin<=sigmax else 0,
                             'start_offset': start_bit_offset,
                             'bit_count': sig_size,
                             'aditional_byte_offset' : additional_byte_offset}
                    ch = Channel(**kargs)
                    ch.name = name
                    gp_channels.append(ch)
                    offset += sig_size

                    # update channel database with component indexes
                    if not name in self.channels_db:
                        self.channels_db[name] = []
                    self.channels_db[name].append((dg_cntr, ch_cntr))

                    # also update record parents
                    parents[ch_cntr] = name, 0

                    ch_cntr += 1

                    # update parent channel dependency referenced channels list
                    parent_dep.referenced_channels.append((ch, gp))

        #channel group
        kargs = {'cycles_nr': cycles_nr,
                 'samples_byte_nr': offset >> 3}
        gp['channel_group'] = ChannelGroup(**kargs)
        gp['channel_group']['ch_nr'] = channel_nr + 1
        gp['size'] = cycles_nr * (offset >> 3)

        #data group
        kargs = {'block_len': DG32_BLOCK_SIZE if self.version in ('3.20', '3.30') else DG31_BLOCK_SIZE}
        gp['data_group'] = DataGroup(**kargs)

        #data block
        types = dtype(types)

        gp['types'] = types
        gp['parents'] = parents

        samples = fromarrays(arrays, dtype=types)
        block = samples.tostring()

        if self.load_measured_data:
            gp['data_location'] = LOCATION_MEMORY
            kargs = {'data': block}
            gp['data_block'] = DataBlock(**kargs)
        else:
            gp['data_location'] = LOCATION_TEMPORARY_FILE
            if self._tempfile is None:
                self._tempfile = TemporaryFile()
            self._tempfile.seek(0, SEEK_END)
            data_address = self._tempfile.tell()
            gp['data_group']['data_block_addr'] = data_address
            self._tempfile.write(block)

        # data group trigger
        gp['trigger'] = [None, None]

    def close(self):
        """ if the MDF was created with load_measured_data=False and new channels
        have been appended, then this must be called just before the object is not
        used anymore to clean-up the temporary file"""
        if self.load_measured_data == False and self._tempfile is not None:
            self._tempfile.close()

    def get(self, name=None, group=None, index=None, raster=None, samples_only=False):
        """Gets channel samples.
        Channel can be specified in two ways:

        * using the first positional argument *name*

            * if there are multiple occurances for this channel then the *group* and *index* arguments can be used to select a specific group.
            * if there are multiple occurances for this channel and either the *group* or *index* arguments is None then a warning is issued

        * using the group number (keyword argument *group*) and the channel number (keyword argument *index*). Use *info* method for group and channel numbers


        If the *raster* keyword argument is not *None* the output is interpolated accordingly

        Parameters
        ----------
        name : string
            name of channel
        group : int
            0-based group index
        index : int
            0-based channel index
        raster : float
            time raster in seconds
        samples_only : bool
            if *True* return only the channel samples as numpy array; if *False* return a *Signal* object

        Returns
        -------
        res : (numpy.array | Signal)
            returns *Signal* if *samples_only*=*False* (default option), otherwise returns numpy.array

        Raises
        ------
        MdfError :

        * if the channel name is not found
        * if the group index is out of range
        * if the channel index is out of range

        """
        if name is None:
            if group is None or index is None:
                raise MdfException('Invalid arguments for "get" methos: must give "name" or, "group" and "index"')
            else:
                gp_nr, ch_nr = group, index
                if gp_nr > len(self.groups) - 1:
                    raise MdfException('Group index out of range')
                if index > len(self.groups[gp_nr]['channels']) - 1:
                    raise MdfException('Channel index out of range')
        else:
            if not name in self.channels_db:
                raise MdfException('Channel "{}" not found'.format(name))
            else:
                if group is None or index is None:
                    gp_nr, ch_nr = self.channels_db[name][0]
                    if len(self.channels_db[name]) > 1:
                        warnings.warn('Multiple occurances for channel "{}". Using first occurance from data group {}. Provide both "group" and "index" arguments to select another data group'.format(name, gp_nr))
                else:
                    for gp_nr, ch_nr in self.channels_db[name]:
                        if gp_nr == group:
                            break
                    else:
                        gp_nr, ch_nr = self.channels_db[name][0]
                        warnings.warn('You have selected group "{}" for channel "{}", but this channel was not found in this group. Using first occurance of "{}" from group "{}"'.format(group, name, name, gp_nr))


        grp = self.groups[gp_nr]
        channel = grp['channels'][ch_nr]
        conversion = grp['channel_conversions'][ch_nr]
        dependecy_block = grp['channel_dependencies'][ch_nr]

        try:
            parents, dtypes = grp['parents'], grp['types']
        except:
            grp['parents'], grp['types'] = self._prepare_record(grp)
            parents, dtypes = grp['parents'], grp['types']
        # get data group record
        if not self.load_measured_data:
            data = self._load_group_data(grp)
            record = fromstring(data, dtype=dtypes)
        else:
            try:
                record = grp['record']
            except:
                record = grp['record'] = fromstring(grp['data_block']['data'], dtype=dtypes)

        # check if this is a channel array
        if dependecy_block:
            referenced_channels = [ch for ch, gp_ in dependecy_block.referenced_channels]
            arrays = [self.get(ch.name, samples_only=True) for ch in referenced_channels]
            types = [ (ch.name, arr.dtype) for ch, arr in zip(referenced_channels, arrays)]
            types = dtype(types)
            vals = fromarrays(arrays, dtype=types)

            info = {'type' : SIGNAL_TYPE_V3_ARRAY}

            if samples_only:
                return vals
            else:
                if conversion:
                    unit = conversion['unit'].decode('latin-1').strip(' \n\t\x00')
                else:
                    unit = ''
                comment = channel['description'].decode('latin-1').strip(' \t\n\x00')

                # get master channel index
                time_ch_nr = self.masters_db[gp_nr]


                time_conv = grp['channel_conversions'][time_ch_nr]
                time_ch = grp['channels'][time_ch_nr]

                t = record[time_ch.name]
                # get timestamps
                time_conv_type = CONVERSION_TYPE_NONE if time_conv is None else time_conv['conversion_type']
                if time_conv_type == CONVERSION_TYPE_LINEAR:
                    time_a = time_conv['a']
                    time_b = time_conv['b']
                    t = t * time_a
                    if time_b:
                        t += time_b
                res = Signal(samples=vals,
                             timestamps=t,
                             unit=unit,
                             name=channel.name,
                             comment=comment,
                             conversion=info)

                if raster and t:
                    tx = linspace(0, t[-1], int(t[-1] / raster))
                    res = res.interp(tx)
                return res
        else:
            # get channel values
            parent, bit_offset = parents[ch_nr]
            vals = record[parent]

            if bit_offset:
                vals = vals >> bit_offset
            bits = channel['bit_count']
            if bits % 8:
                vals = vals & ((1<<bits) - 1)

            info = None

            conversion_type = CONVERSION_TYPE_NONE if conversion is None else conversion['conversion_type']

            if conversion_type == CONVERSION_TYPE_NONE:
                # is it a Byte Array?
                if channel['data_type'] == DATA_TYPE_BYTEARRAY:
                    vals = vals.tostring()
                    size = max(bits>>3, 1)
                    cols = size
                    lines = len(vals) // cols

                    vals = frombuffer(vals, dtype=uint8).reshape((lines, cols))

                    info = {'type': SIGNAL_TYPE_V3_BYTEARRAY}

            elif conversion_type == CONVERSION_TYPE_LINEAR:
                a = conversion['a']
                b = conversion['b']
                if (a, b) == (1, 0):
                    size = max(bits>>3, 1)
                    ch_fmt = get_fmt(channel['data_type'], size)
                    if not vals.dtype == ch_fmt:
                        vals = vals.astype(ch_fmt)
                else:
                    vals = vals * a
                    if b:
                        vals += b

            elif conversion_type in (CONVERSION_TYPE_TABI, CONVERSION_TYPE_TABX):
                nr = conversion['ref_param_nr']
                raw = array([conversion['raw_{}'.format(i)] for i in range(nr)])
                phys = array([conversion['phys_{}'.format(i)] for i in range(nr)])
                if conversion_type == CONVERSION_TYPE_TABI:
                    vals = interp(vals, raw, phys)
                else:
                    idx = searchsorted(raw, vals)
                    idx = clip(idx, 0, len(raw) - 1)
                    vals = phys[idx]

            elif conversion_type == CONVERSION_TYPE_VTAB:
                nr = conversion['ref_param_nr']
                raw = array([conversion['param_val_{}'.format(i)] for i in range(nr)])
                phys = array([conversion['text_{}'.format(i)] for i in range(nr)])
                info = {'raw': raw, 'phys': phys, 'type': SIGNAL_TYPE_V3_VTAB}

            elif conversion_type == CONVERSION_TYPE_VTABR:
                nr = conversion['ref_param_nr']

                texts = array([grp['texts']['conversion_tab'][ch_nr].get('text_{}'.format(i), {}).get('text', b'') for i in range(nr)])
                lower = array([conversion['lower_{}'.format(i)] for i in range(nr)])
                upper = array([conversion['upper_{}'.format(i)] for i in range(nr)])
                info = {'lower': lower, 'upper': upper, 'phys': texts, 'type': SIGNAL_TYPE_V3_VTABR}

            elif conversion_type in (CONVERSION_TYPE_EXPO, CONVERSION_TYPE_LOGH):
                func = log if conversion_type == CONVERSION_TYPE_EXPO else exp
                P1 = conversion['P1']
                P2 = conversion['P2']
                P3 = conversion['P3']
                P4 = conversion['P4']
                P5 = conversion['P5']
                P6 = conversion['P6']
                P7 = conversion['P7']
                if P4 == 0:
                    vals = func(((vals - P7) * P6 - P3) / P1) / P2
                elif P1 == 0:
                    vals = func((P3 / (vals - P7) - P6) / P4) / P5
                else:
                    raise ValueError('wrong conversion type {}'.format(conversion_type))

            elif conversion_type == CONVERSION_TYPE_RAT:
                P1 = conversion['P1']
                P2 = conversion['P2']
                P3 = conversion['P3']
                P4 = conversion['P4']
                P5 = conversion['P5']
                P6 = conversion['P6']
                X = vals
                vals = evaluate('(P1 * X**2 + P2 * X + P3) / (P4 * X**2 + P5 * X + P6)')

            elif conversion_type == CONVERSION_TYPE_POLY:
                P1 = conversion['P1']
                P2 = conversion['P2']
                P3 = conversion['P3']
                P4 = conversion['P4']
                P5 = conversion['P5']
                P6 = conversion['P6']
                X = vals
                vals = evaluate('(P2 - (P4 * (X - P5 -P6))) / (P3* (X - P5 - P6) - P1)')

            elif conversion_type == CONVERSION_TYPE_FORMULA:
                formula = conversion['formula'].decode('latin-1').strip(' \n\t\x00')
                X1 = vals
                vals = evaluate(formula)

            if samples_only:
                return vals
            else:
                if conversion:
                    unit = conversion['unit'].decode('latin-1').strip(' \n\t\x00')
                else:
                    unit = ''
                comment = channel['description'].decode('latin-1').strip(' \t\n\x00')

                # get master channel index
                time_ch_nr = self.masters_db[gp_nr]

                if time_ch_nr == ch_nr:
                    res = Signal(samples=vals.copy(),
                                 timestamps=vals,
                                 unit=unit,
                                 name=channel.name,
                                 comment=comment)
                else:
                    time_conv = grp['channel_conversions'][time_ch_nr]
                    time_ch = grp['channels'][time_ch_nr]
                    t = record[time_ch.name]
                    # get timestamps
                    time_conv_type = CONVERSION_TYPE_NONE if time_conv is None else time_conv['conversion_type']
                    if time_conv_type == CONVERSION_TYPE_LINEAR:
                        time_a = time_conv['a']
                        time_b = time_conv['b']
                        t = t * time_a
                        if time_b:
                            t += time_b
                    res = Signal(samples=vals,
                                 timestamps=t,
                                 unit=unit,
                                 name=channel.name,
                                 comment=comment,
                                 conversion=info)

                if raster and t:
                    tx = linspace(0, t[-1], int(t[-1] / raster))
                    res = res.interp(tx)
                return res

    def iter_get_triggers(self):
        """ generator that yields triggers

        Returns
        -------
        trigger_info : dict
            trigger information with the following keys:

                * comment : trigger comment
                * time : trigger time
                * pre_time : trigger pre time
                * post_time : trigger post time
                * index : trigger index
                * group : data group index of trigger
        """
        for i, gp in enumerate(self.groups):
            trigger, trigger_text = gp['trigger']
            if trigger:
                if trigger_text:
                    comment = trigger_text.text_str
                else:
                    comment = ''

                for j in range(trigger['trigger_events_nr']):
                    trigger_info = {'comment': comment,
                                    'index' : j,
                                    'group': i,
                                    'time' : trigger['trigger_{}_time'.format(j)],
                                    'pre_time' : trigger['trigger_{}_pretime'.format(j)],
                                    'post_time' : trigger['trigger_{}_posttime'.format(j)]}
                    yield trigger_info

    def info(self):
        """get MDF information as a dict

        Examples
        --------
        >>> mdf = MDF3('test.mdf')
        >>> mdf.info()

        """
        info = {}
        info['version'] = self.identification['version_str'].strip(b'\x00').decode('latin-1')
        info['author'] = self.header['author'].strip(b'\x00').decode('latin-1')
        info['organization'] = self.header['organization'].strip(b'\x00').decode('latin-1')
        info['project'] = self.header['project'].strip(b'\x00').decode('latin-1')
        info['subject'] = self.header['subject'].strip(b'\x00').decode('latin-1')
        info['groups'] = len(self.groups)
        for i, gp in enumerate(self.groups):
            inf = {}
            info['group {}'.format(i)] = inf
            inf['cycles'] = gp['channel_group']['cycles_nr']
            inf['channels count'] = len(gp['channels'])
            for j, ch in enumerate(gp['channels']):
                inf['channel {}'.format(j)] = (ch.name, ch['channel_type'])

        return info

    def remove(self, group=None, name=None):
        """Remove data group. Use *group* or *name* keyword arguments to identify the group's index. *group* has priority

        Parameters
        ----------
        name : string
            name of the channel inside the data group to be removed
        group : int
            data group index to be removed

        Examples
        --------
        >>> mdf = MDF3('test.mdf')
        >>> mdf.remove(group=3)
        >>> mdf.remove(name='VehicleSpeed')

        """
        if group:
            if 0 <= group <= len(self.groups):
                idx = group
            else:
                print('Group index "{}" not in valid range[0..{}]'.format(group, len(self.groups)))
                return
        elif name:
            if name in self.channels_db:
                idx = self.channels_db[name][1]
            else:
                print('Channel name "{}" not found in the measurement'.format(name))
                return
        else:
            print('Must specify a valid group or name argument')
            return
        self.groups.pop(idx)

    def save(self, dst='', overwrite=False):
        """Save MDF to *dst*. If *dst* is not provided the the destination file name is
        the MDF name. If overwrite is *True* then the destination file is overwritten,
        otherwise the file name is appened with '_xx', were 'xx' is the first conter that produces a new
        file name (that does not already exist in the filesystem)

        Parameters
        ----------
        dst : str
            destination file name, Default ''
        overwrite : bool
            overwrite flag, default *False*

        """

        if self.file_history is None:
            self.file_history = TextBlock.from_text('''<FHcomment>
<TX>created</TX>
<tool_id>asammdf</tool_id>
<tool_vendor> </tool_vendor>
<tool_version>2.4.1</tool_version>
</FHcomment>''')
        else:
            text = self.file_history['text'] + '\n{}: updated byt Python script'.format(time.asctime()).encode('latin-1')
            self.file_history = TextBlock.from_text(text)

        if self.name is None and dst == '':
            raise MdfException('Must specify a destination file name for MDF created from scratch')

        dst = dst if dst else self.name
        if overwrite == False:
            if os.path.isfile(dst):
                cntr = 0
                while True:
                    name = os.path.splitext(dst)[0] + '_{}.mdf'.format(cntr)
                    if not os.path.isfile(name):
                        break
                    else:
                        cntr += 1
                warnings.warn('Destination file "{}" already exists and "overwrite" is False. Saving MDF file as "{}"'.format(dst, name))
                dst = name

        # all MDF blocks are appended to the blocks list in the order in which
        # they will be written to disk. While creating this list, all the relevant
        # block links are updated so that once all blocks have been added to the list
        # they can simply be written (using the bytes protocol).
        # DataGroup blocks are written first after the identification and header blocks.
        # When load_measured_data=False we need to restore the original data block addresses
        # within the data group block. This is needed to allow further work with the object
        # after the save method call (eq. new calls to get method). Since the data group blocks
        # are written first, it is safe to restor the original links when the data blocks
        # are written. For lado_measured_data=False, the blocks list will contain a tuple
        # instead of a DataBlock instance; the tuple will have the reference to the
        # data group object and the original link to the data block in the soource MDF file.

        with open(dst, 'wb') as dst:
            #store unique texts and their addresses
            defined_texts = {}
            # list of all blocks
            blocks = []

            address = 0

            blocks.append(self.identification)
            address += ID_BLOCK_SIZE

            blocks.append(self.header)
            address += self.header['block_len']

            self.file_history.address = address
            blocks.append(self.file_history)
            address += self.file_history['block_len']

            # DataGroup
            # put them first in the block list so they will be written first to disk
            # this way, in case of load_measured_data=False, we can safely restore
            # the original data block address
            for gp in self.groups:
                dg = gp['data_group']
                blocks.append(dg)
                dg.address = address
                address += dg['block_len']
            for i, dg in enumerate(self.groups[:-1]):
                dg['data_group']['next_dg_addr'] = self.groups[i+1]['data_group'].address
            self.groups[-1]['data_group']['next_dg_addr'] = 0

            for index, gp in enumerate(self.groups):
                gp_texts = gp['texts']

                # Texts
                for _, item_list in gp_texts.items():
                    for my_dict in item_list:
                        for key in my_dict:
                            #text blocks can be shared
                            if my_dict[key].text_str in defined_texts:
                                my_dict[key].address = defined_texts[my_dict[key].text_str]
                            else:
                                defined_texts[my_dict[key].text_str] = address
                                my_dict[key].address = address
                                blocks.append(my_dict[key])
                                address += my_dict[key]['block_len']

                # ChannelConversions
                cc = gp['channel_conversions']
                for i, conv in enumerate(cc):
                    if conv:
                        conv.address = address
                        if conv['conversion_type'] == CONVERSION_TYPE_VTABR:
                            for key, item in gp_texts['conversion_tab'][i].items():
                                conv[key] = item.address

                        blocks.append(conv)
                        address += conv['block_len']

                # Channel Extension
                cs = gp['channel_extensions']
                for source in cs:
                    if source:
                        source.address = address
                        blocks.append(source)
                        address += source['block_len']

                # Channel Dependency
                cd = gp['channel_dependencies']
                for dep in cd:
                    if dep:
                        dep.address = address
                        blocks.append(dep)
                        address += dep['block_len']

                # Channels
                ch_texts = gp_texts['channels']
                for i, channel in enumerate(gp['channels']):
                    channel.address = address
                    blocks.append(channel)
                    address += CN_BLOCK_SIZE

                    for key in ('long_name_addr', 'comment_addr', 'display_name_addr'):
                        channel_texts = ch_texts[i]
                        if key in channel_texts:
                            channel[key] = channel_texts[key].address
                        else:
                            channel[key] = 0
                    channel['conversion_addr'] = cc[i].address if cc[i] else 0
                    channel['source_depend_addr'] = cs[i].address if cs[i] else 0
                    if cd[i]:
                        channel['ch_depend_addr'] = cd[i].address
                    else:
                        channel['ch_depend_addr'] = 0

                for channel, next_channel in pair(gp['channels']):
                    channel['next_ch_addr'] = next_channel.address
                next_channel['next_ch_addr'] = 0

                # ChannelGroup
                cg = gp['channel_group']
                cg.address = address
                blocks.append(cg)
                address += cg['block_len']

                cg['first_ch_addr'] = gp['channels'][0].address
                cg['next_cg_addr'] = 0
                if 'comment_addr' in gp['texts']['channel_group'][0]:
                    cg['comment_addr'] = gp_texts['channel_group'][0]['comment_addr'].address

                # TriggerBLock
                trigger, trigger_text = gp['trigger']
                if trigger:
                    if trigger_text:
                        trigger_text.address = address
                        blocks.append(trigger_text)
                        address += trigger_text['block_len']
                        trigger['comment_addr'] = trigger_text.address
                    else:
                        trigger['comment_addr'] = 0

                    trigger.address = address
                    blocks.append(trigger)
                    address += trigger['block_len']

                # DataBlock
                original_data_addr = gp['data_group']['data_block_addr']
                gp['data_group']['data_block_addr'] = address if gp['size'] else 0
                address += gp['size']
                if self.load_measured_data:
                    blocks.append(gp['data_block'])
                else:
                    # trying to call bytes([gp, address]) will result in an exception
                    # that be used as a flag for non existing data block in case
                    # of load_measured_data=False, the address is the actual address
                    # of the data group's data within the original file
                    blocks.append([gp, original_data_addr])

            # update referenced channels addresses within the channel dependecies
            for gp in self.groups:
                for dep in gp['channel_dependencies']:
                    if dep:
                        for i, (ch, grp) in enumerate(dep.referenced_channels):
                            dep['ch_{}'.format(i)] = ch.address
                            dep['cg_{}'.format(i)] = grp['channel_group'].address
                            dep['dg_{}'.format(i)] = grp['data_group'].address

            # DataGroup
            for gp in self.groups:
                gp['data_group']['first_cg_addr'] = gp['channel_group'].address
                gp['data_group']['trigger_addr'] = gp['trigger'][0].address if gp['trigger'][0] else 0

            if self.groups:
                self.header['first_dg_addr'] = self.groups[0]['data_group'].address
                self.header['dg_nr'] = len(self.groups)
                self.header['comment_addr'] = self.file_history.address
                self.header['program_addr'] = 0

            write = dst.write
            for block in blocks:
                try:
                    write(bytes(block))
                except:
                    # this will only be executed for data blocks when load_measured_data=False
                    gp, address = block
                    # restore data block address from original file so that
                    # future calls to get will still work after the save
                    gp['data_group']['data_block_addr'] = address
                    data = self._load_group_data(gp)
                    write(data)


if __name__ == '__main__':
    pass



"""
Signal constants
"""

SIGNAL_TYPE_V3_VTAB = 0
SIGNAL_TYPE_V3_VTABR = 1
SIGNAL_TYPE_V3_BYTEARRAY = 2
SIGNAL_TYPE_V3_ARRAY = 3

SIGNAL_TYPE_V4_VTAB = 4
SIGNAL_TYPE_V4_VTABR = 5
SIGNAL_TYPE_V4_BYTEARRAY = 6
SIGNAL_TYPE_V4_ARRAY = 7
SIGNAL_TYPE_V4_TTAB = 8
SIGNAL_TYPE_V4_TRANS = 9
SIGNAL_TYPE_V4_CANOPENDATE = 10
SIGNAL_TYPE_V4_CANOPENTIME = 11



from __future__ import print_function, division
import sys
PYVERSION = sys.version_info[0]
PYVERSION_MAJOR = sys.version_info[0] * 10 + sys.version_info[1]

import time
import warnings

from hashlib import md5
from struct import unpack, pack, unpack_from

from zlib import compress, decompress

import numpy as np

from .v4constants import *


__all__ = ['AttachmentBlock',
           'Channel',
           'ChannelArrayBlock',
           'ChannelGroup',
           'ChannelConversion',
           'DataBlock',
           'DataZippedBlock',
           'FileIdentificationBlock',
           'HeaderBlock',
           'HeaderList',
           'DataList',
           'DataGroup',
           'FileHistory',
           'SignalDataBlock',
           'SourceInformation',
           'TextBlock']


class AttachmentBlock(dict):
    """ ATBLOCK class

    When adding new attachments only embedded attachemnts are allowed, with keyword argument *data* of type bytes"""
    def __init__(self, **kargs):
        super(AttachmentBlock, self).__init__()

        try:
            self.address = address = kargs['address']
            stream = kargs['file_stream']
            stream.seek(address, SEEK_START)

            (self['id'],
             self['reserved0'],
             self['block_len'],
             self['links_nr'],
             self['next_at_addr'],
             self['file_name_addr'],
             self['mime_addr'],
             self['comment_addr'],
             self['flags'],
             self['creator_index'],
             self['reserved1'],
             self['md5_sum'],
             self['original_size'],
             self['embedded_size']) = unpack(FMT_AT_COMMON, stream.read(AT_COMMON_SIZE))

            self['embedded_data'] = stream.read(self['embedded_size'])

        except KeyError:

            data = kargs['data']
            size = len(data)
            compression = kargs.get('compression', False)

            if compression:
                data = compress(data)
                original_size = size
                size = len(data)
                self['id'] = b'##AT'
                self['reserved0'] = 0
                self['block_len'] = AT_COMMON_SIZE + size
                self['links_nr'] = 4
                self['next_at_addr'] = 0
                self['file_name_addr'] = 0
                self['mime_addr'] = 0
                self['comment_addr'] = 0
                self['flags'] = FLAG_AT_EMBEDDED | FLAG_AT_MD5_VALID | FLAG_AT_COMPRESSED_EMBEDDED
                self['creator_index'] = 0
                self['reserved1'] = 0
                md5_worker = md5()
                md5_worker.update(data)
                self['md5_sum'] = md5_worker.digest()
                self['original_size'] = original_size
                self['embedded_size'] = size
                self['embedded_data'] = data
            else:
                self['id'] = b'##AT'
                self['reserved0'] = 0
                self['block_len'] = AT_COMMON_SIZE + size
                self['links_nr'] = 4
                self['next_at_addr'] = 0
                self['file_name_addr'] = 0
                self['mime_addr'] = 0
                self['comment_addr'] = 0
                self['flags'] = FLAG_AT_EMBEDDED | FLAG_AT_MD5_VALID
                self['creator_index'] = 0
                self['reserved1'] = 0
                md5_worker = md5()
                md5_worker.update(data)
                self['md5_sum'] = md5_worker.digest()
                self['original_size'] = size
                self['embedded_size'] = size
                self['embedded_data'] = data

    def extract(self):
        if self['flags'] & FLAG_AT_EMBEDDED:
            if self['flags'] & FLAG_AT_COMPRESSED_EMBEDDED:
                data = decompress(self['embedded_data'])
            else:
                data = self['embedded_data']
            if self['flags'] & FLAG_AT_MD5_VALID:
                md5_worker = md5()
                md5_worker.update(data)
                md5_sum = md5_worker.digest()
                if self['md5_sum'] == md5_sum:
                    return data
                else:
                    warnings.warn('ATBLOCK md5sum="{}" and embedded data md5sum="{}"'.format(self['md5_sum'], md5_sum))
        else:
            warnings.warn('extarnal attachments not supported')

    def __bytes__(self):
        fmt = FMT_AT_COMMON + '{}s'.format(self['embedded_size'])
        if PYVERSION_MAJOR >= 36:
            return pack(fmt, *self.values())
        else:
            return pack(fmt, *[self[key] for key in KEYS_AT_BLOCK])

class Channel(dict):
    """ CNBLOCK class"""
    def __init__(self, **kargs):
        super(Channel, self).__init__()

        self.name = ''
        self.cn_template = False
        self.cn_template_size = 0

        try:
            self.address = address = kargs['address']
            stream = kargs['file_stream']
            stream.seek(address, SEEK_START)

            (self['id'],
             self['reserved0'],
             self['block_len'],
             self['links_nr'],
             self['next_ch_addr'],
             self['component_addr'],
             self['name_addr'],
             self['source_addr'],
             self['conversion_addr'],
             self['data_block_addr'],
             self['unit_addr'],
             self['comment_addr'],
             self['channel_type'],
             self['sync_type'],
             self['data_type'],
             self['bit_offset'],
             self['byte_offset'],
             self['bit_count'],
             self['flags'],
             self['pos_invalidation_bit'],
             self['precision'],
             self['reserved1'],
             self['attachment_nr'],
             self['min_raw_value'],
             self['max_raw_value'],
             self['lower_limit'],
             self['upper_limit'],
             self['lower_ext_limit'],
             self['upper_ext_limit']) = unpack(FMT_CHANNEL, stream.read(CN_BLOCK_SIZE))

        except KeyError:

            self.address = 0

            self['id'] = b'##CN'
            self['reserved0'] = 0
            self['block_len'] = CN_BLOCK_SIZE
            self['links_nr'] = 8
            self['next_ch_addr'] = 0
            self['component_addr'] = 0
            self['name_addr'] = 0
            self['source_addr'] = 0
            self['conversion_addr'] = 0
            self['data_block_addr'] = 0
            self['unit_addr'] = 0
            self['comment_addr'] = 0
            self['channel_type'] = kargs.get('channel_type', 0)
            self['sync_type'] = kargs.get('sync_type', 0)
            self['data_type'] = kargs.get('data_type', 0)
            self['bit_offset'] = kargs.get('bit_offset', 0)
            self['byte_offset'] = kargs.get('byte_offset', 8)
            self['bit_count'] = kargs.get('bit_count', 8)
            self['flags'] = 28
            self['pos_invalidation_bit'] = 0
            self['precision'] = 3
            self['reserved1'] = 0
            self['attachment_nr'] = 0
            self['min_raw_value'] = kargs.get('min_raw_value', 0)
            self['max_raw_value'] = kargs.get('max_raw_value', 0)
            self['lower_limit'] = kargs.get('lower_limit', 0)
            self['upper_limit'] = kargs.get('upper_limit', 100)
            self['lower_ext_limit'] = kargs.get('lower_ext_limit', 0)
            self['upper_ext_limit'] = kargs.get('upper_ext_limit', 0)

    def __bytes__(self):
        if PYVERSION_MAJOR >= 36:
            return pack(FMT_CHANNEL, *self.values())
        else:
            return pack(FMT_CHANNEL, *[self[key] for key in KEYS_CHANNEL])

    def __lt__(self, other):
        if self['byte_offset'] < other['byte_offset']:
            return 1
        elif self['byte_offset'] == other['byte_offset']:
            if self['bit_count'] + self['bit_offset'] > other['bit_count'] + other['bit_offset']:
                return 1
            elif self['bit_count'] + self['bit_offset'] == other['bit_count'] + other['bit_offset']:
                if self['bit_offset'] < other['bit_offset']:
                    return 1
                else:
                    return 0
            else:
                return 0
        else:
            return 0


class ChannelArrayBlock(dict):
    """CABLOCK class"""
    def __init__(self, **kargs):
        super(ChannelArrayBlock, self).__init__()

        try:
            self.address = address = kargs['address']
            stream = kargs['file_stream']
            stream.seek(address, SEEK_START)

            (self['id'],
             self['reserved0'],
             self['block_len'],
             self['links_nr']) = unpack('<4sI2Q', stream.read(24))

            nr = self['links_nr']
            links = unpack('<{}Q'.format(nr), stream.read(8*nr))
            for i in range(nr):
                self['link_{}'.format(i)] = links[i]

            (self['ca_type'],
             self['storage'],
             self['dims'],
             self['flags'],
             self['byte_offset_base'],
             self['invalidation_bit_base'],
             self['dim_size']) = unpack('<2BHIiIQ', stream.read(24))

        except KeyError:
            pass

    def __bytes__(self):
        return b''


class ChannelGroup(dict):
    """CGBLOCK class"""
    def __init__(self, **kargs):
        super(ChannelGroup, self).__init__()

        try:
            self.address = address = kargs['address']
            stream = kargs['file_stream']
            stream.seek(address, SEEK_START)

            (self['id'],
             self['reserved0'],
             self['block_len'],
             self['links_nr'],
             self['next_cg_addr'],
             self['first_ch_addr'],
             self['acq_name_addr'],
             self['acq_source_addr'],
             self['first_sample_reduction_addr'],
             self['comment_addr'],
             self['record_id'],
             self['cycles_nr'],
             self['flags'],
             self['path_separator'],
             self['reserved1'],
             self['samples_byte_nr'],
             self['invalidation_bytes_nr']) = unpack(FMT_CHANNEL_GROUP, stream.read(CG_BLOCK_SIZE))

        except KeyError:
            self.address = 0
            self['id'] = kargs.get('id', '##CG'.encode('utf-8'))
            self['reserved0'] = kargs.get('reserved0', 0)
            self['block_len'] = kargs.get('block_len', CG_BLOCK_SIZE)
            self['links_nr'] = kargs.get('links_nr', 6)
            self['next_cg_addr'] = kargs.get('next_cg_addr', 0)
            self['first_ch_addr'] = kargs.get('first_ch_addr', 0)
            self['acq_name_addr'] = kargs.get('acq_name_addr', 0)
            self['acq_source_addr'] = kargs.get('acq_source_addr', 0)
            self['first_sample_reduction_addr'] = kargs.get('first_sample_reduction_addr', 0)
            self['comment_addr'] = kargs.get('comment_addr', 0)
            self['record_id'] = kargs.get('record_id', 1)
            self['cycles_nr'] = kargs.get('cycles_nr', 0)
            self['flags'] = kargs.get('flags', 0)
            self['path_separator'] = kargs.get('path_separator', 0)
            self['reserved1'] = kargs.get('reserved1', 0)
            self['samples_byte_nr'] = kargs.get('samples_byte_nr', 0)
            self['invalidation_bytes_nr'] = kargs.get('invalidation_bytes_nr', 0)

    def __bytes__(self):
        if PYVERSION_MAJOR >= 36:
            return pack(FMT_CHANNEL_GROUP, *self.values())
        else:
            return pack(FMT_CHANNEL_GROUP, *[self[key] for key in KEYS_CHANNEL_GROUP])


class ChannelConversion(dict):
    """CCBLOCK class"""
    def __init__(self, **kargs):
        super(ChannelConversion, self).__init__()

        try:
            self.address = address = kargs['address']
            stream = kargs['file_stream']
            stream.seek(address, SEEK_START)

            (self['id'],
             self['reserved0'],
             self['block_len'],
             self['links_nr']) = unpack(FMT_COMMON, stream.read(COMMON_SIZE))

            block = stream.read(self['block_len'] - COMMON_SIZE)

            conv = unpack_from('B', block, self['links_nr'] * 8)[0]

            if conv == CONVERSION_TYPE_NON:
                (self['name_addr'],
                 self['unit_addr'],
                 self['comment_addr'],
                 self['inv_conv_addr'],
                 self['conversion_type'],
                 self['precision'],
                 self['flags'],
                 self['ref_param_nr'],
                 self['val_param_nr'],
                 self['min_phy_value'],
                 self['max_phy_value']) = unpack('<4Q2B3H2d', block)
                
            elif conv == CONVERSION_TYPE_LIN:
                (self['name_addr'],
                 self['unit_addr'],
                 self['comment_addr'],
                 self['inv_conv_addr'],
                 self['conversion_type'],
                 self['precision'],
                 self['flags'],
                 self['ref_param_nr'],
                 self['val_param_nr'],
                 self['min_phy_value'],
                 self['max_phy_value'],
                 self['b'],
                 self['a']) = unpack('<4Q2B3H4d', block)
                
            elif conv == CONVERSION_TYPE_RAT:
                (self['name_addr'],
                 self['unit_addr'],
                 self['comment_addr'],
                 self['inv_conv_addr'],
                 self['conversion_type'],
                 self['precision'],
                 self['flags'],
                 self['ref_param_nr'],
                 self['val_param_nr'],
                 self['min_phy_value'],
                 self['max_phy_value'],
                 self['P1'],
                 self['P2'],
                 self['P3'],
                 self['P4'],
                 self['P5'],
                 self['P6']) = unpack('<4Q2B3H8d', block)

            elif conv == CONVERSION_TYPE_ALG:
                (self['name_addr'],
                 self['unit_addr'],
                 self['comment_addr'],
                 self['inv_conv_addr'],
                 self['formula_addr'],
                 self['conversion_type'],
                 self['precision'],
                 self['flags'],
                 self['ref_param_nr'],
                 self['val_param_nr'],
                 self['min_phy_value'],
                 self['max_phy_value']) = unpack('<5Q2B3H2d', block)

            elif conv in (CONVERSION_TYPE_TABI, CONVERSION_TYPE_TAB):
                (self['name_addr'],
                 self['unit_addr'],
                 self['comment_addr'],
                 self['inv_conv_addr'],
                 self['conversion_type'],
                 self['precision'],
                 self['flags'],
                 self['ref_param_nr'],
                 self['val_param_nr'],
                 self['min_phy_value'],
                 self['max_phy_value']) = unpack_from('<4Q2B3H2d', block)

                nr = self['val_param_nr']
                values = unpack('<{}d'.format(nr), block[56:])
                for i in range(nr // 2):
                    (self['raw_{}'.format(i)],
                     self['phys_{}'.format(i)]) = values[i*2], values[2*i+1]

            elif conv == CONVERSION_TYPE_RTAB:
                (self['name_addr'],
                 self['unit_addr'],
                 self['comment_addr'],
                 self['inv_conv_addr'],
                 self['conversion_type'],
                 self['precision'],
                 self['flags'],
                 self['ref_param_nr'],
                 self['val_param_nr'],
                 self['min_phy_value'],
                 self['max_phy_value']) = unpack_from('<4Q2B3H2d', block)
                nr = self['val_param_nr']
                values = unpack('<{}d'.format(nr), block[56:])
                for i in range((nr - 1) // 3):
                    (self['lower_{}'.format(i)],
                     self['upper_{}'.format(i)],
                     self['phys_{}'.format(i)]) = values[i*3], values[3*i+1], values[3*i+2]
                self['default'] = unpack('<d', block[-8:])[0]

            elif conv == CONVERSION_TYPE_TABX:
                (self['name_addr'],
                 self['unit_addr'],
                 self['comment_addr'],
                 self['inv_conv_addr']) = unpack_from('<4Q', block)

                links_nr = self['links_nr'] - 4

                links = unpack_from('<{}Q'.format(links_nr), block, 32)
                for i, link in enumerate(links[:-1]):
                    self['text_{}'.format(i)] = link
                self['default_addr'] = links[-1]

                (self['conversion_type'],
                 self['precision'],
                 self['flags'],
                 self['ref_param_nr'],
                 self['val_param_nr'],
                 self['min_phy_value'],
                 self['max_phy_value']) = unpack_from('<2B3H2d', block, 32 + links_nr * 8)

                values = unpack_from('<{}d'.format(links_nr - 1), block, 32 + links_nr * 8 + 24)
                for i, val in enumerate(values):
                    self['val_{}'.format(i)] = val

            elif conv == CONVERSION_TYPE_RTABX:
                (self['name_addr'],
                 self['unit_addr'],
                 self['comment_addr'],
                 self['inv_conv_addr']) = unpack_from('<4Q', block)

                links_nr = self['links_nr'] - 4

                links = unpack_from('<{}Q'.format(links_nr), block, 32)
                for i, link in enumerate(links[:-1]):
                    self['text_{}'.format(i)] = link
                self['default_addr'] = links[-1]

                (self['conversion_type'],
                 self['precision'],
                 self['flags'],
                 self['ref_param_nr'],
                 self['val_param_nr'],
                 self['min_phy_value'],
                 self['max_phy_value']) = unpack_from('<2B3H2d', block, 32 + links_nr * 8)

                values = unpack_from('<{}d'.format( (links_nr - 1) * 2 ), block, 32 + links_nr * 8 + 24)
                for i in range(self['val_param_nr'] // 2):
                    j = 2 * i
                    self['lower_{}'.format(i)] = values[j]
                    self['upper_{}'.format(i)] = values[j+1]

            elif conv == CONVERSION_TYPE_TTAB:
                (self['name_addr'],
                 self['unit_addr'],
                 self['comment_addr'],
                 self['inv_conv_addr']) = unpack_from('<4Q', block)

                links_nr = self['links_nr'] - 4

                links = unpack_from('<{}Q'.format(links_nr), block, 32)
                for i, link in enumerate(links[:-1]):
                    self['text_{}'.format(i)] = link

                (self['conversion_type'],
                 self['precision'],
                 self['flags'],
                 self['ref_param_nr'],
                 self['val_param_nr'],
                 self['min_phy_value'],
                 self['max_phy_value']) = unpack_from('<2B3H2d', block, 32 + links_nr * 8)

                values = unpack_from('<{}d'.format(self['val_param_nr']), block, 32 + links_nr * 8 + 24)
                for i, val in enumerate(values[:-1]):
                    self['val_{}'.format(i)] = val
                self['val_default'] = values[-1]

            elif conv == CONVERSION_TYPE_TRANS:
                (self['name_addr'],
                 self['unit_addr'],
                 self['comment_addr'],
                 self['inv_conv_addr']) = unpack_from('<4Q', block)

                links_nr = self['links_nr'] - 4

                links = unpack_from('<{}Q'.format(links_nr), block, 32)

                for i in range((links_nr -1) // 2):
                    j = 2 * i
                    self['input_{}_addr'.format(i)] = links[j]
                    self['output_{}_addr'.format(i)] = links[j+1]
                self['default_addr'] = links[-1]

                (self['conversion_type'],
                 self['precision'],
                 self['flags'],
                 self['ref_param_nr'],
                 self['val_param_nr'],
                 self['min_phy_value'],
                 self['max_phy_value']) = unpack_from('<2B3H2d', block, 32 + links_nr * 8)

        except KeyError:

            self.address = 0
            self['id'] = '##CC'.encode('utf-8')
            self['reserved0'] = 0

            if kargs['conversion_type'] == CONVERSION_TYPE_NON:
                self['block_len'] = CC_NONE_BLOCK_SIZE
                self['links_nr'] = 4
                self['name_addr'] = 0
                self['unit_addr'] = 0
                self['comment_addr'] = 0
                self['inv_conv_addr'] = 0
                self['conversion_type'] = CONVERSION_TYPE_NON
                self['precision'] = 1
                self['flags'] = 0
                self['ref_param_nr'] = 0
                self['val_param_nr'] = 0
                self['min_phy_value'] = kargs.get('min_phy_value', 0)
                self['max_phy_value'] = kargs.get('max_phy_value', 0)
                
            elif kargs['conversion_type'] == CONVERSION_TYPE_LIN:
                self['block_len'] = kargs.get('block_len', CC_LIN_BLOCK_SIZE)
                self['links_nr'] = kargs.get('links_nr', 4)
                self['name_addr'] = kargs.get('name_addr', 0)
                self['unit_addr'] = kargs.get('unit_addr', 0)
                self['comment_addr'] = kargs.get('comment_addr', 0)
                self['inv_conv_addr'] = kargs.get('inv_conv_addr', 0)
                self['conversion_type'] = CONVERSION_TYPE_LIN
                self['precision'] = kargs.get('precision', 1)
                self['flags'] = kargs.get('flags', 0)
                self['ref_param_nr'] = kargs.get('ref_param_nr', 0)
                self['val_param_nr'] = kargs.get('val_param_nr', 2)
                self['min_phy_value'] = kargs.get('min_phy_value', 0)
                self['max_phy_value'] = kargs.get('max_phy_value', 0)
                self['b'] = kargs.get('b', 0)
                self['a'] = kargs.get('a', 1)
                
            elif kargs['conversion_type'] == CONVERSION_TYPE_ALG:
                self['block_len'] = kargs.get('block_len', CC_ALG_BLOCK_SIZE)
                self['links_nr'] = kargs.get('links_nr', 5)
                self['name_addr'] = kargs.get('name_addr', 0)
                self['unit_addr'] = kargs.get('unit_addr', 0)
                self['comment_addr'] = kargs.get('comment_addr', 0)
                self['inv_conv_addr'] = kargs.get('inv_conv_addr', 0)
                self['conv_text_addr'] = kargs.get('conv_text_addr', 0)
                self['conversion_type'] = CONVERSION_TYPE_ALG
                self['precision'] = kargs.get('precision', 1)
                self['flags'] = kargs.get('flags', 0)
                self['ref_param_nr'] = kargs.get('ref_param_nr', 1)
                self['val_param_nr'] = kargs.get('val_param_nr', 0)
                self['min_phy_value'] = kargs.get('min_phy_value', 0)
                self['max_phy_value'] = kargs.get('max_phy_value', 0)
                
            elif kargs['conversion_type'] == CONVERSION_TYPE_TABX:
                self['block_len'] = ((kargs['links_nr'] - 5) * 8 * 2) + 88
                self['links_nr'] = kargs['links_nr']
                self['name_addr'] = kargs.get('name_addr', 0)
                self['unit_addr'] = kargs.get('unit_addr', 0)
                self['comment_addr'] = kargs.get('comment_addr', 0)
                self['inv_conv_addr'] = kargs.get('inv_conv_addr', 0)
                for i in range(kargs['links_nr'] - 5):
                    self['text_{}'.format(i)] = 0
                self['default_addr'] = kargs.get('default_addr', 0)
                self['conversion_type'] = CONVERSION_TYPE_TABX
                self['precision'] = kargs.get('precision', 0)
                self['flags'] = kargs.get('flags', 0)
                self['ref_param_nr'] = kargs.get('ref_param_nr', kargs['links_nr'] - 4)
                self['val_param_nr'] = kargs.get('val_param_nr', kargs['links_nr'] - 5)
                self['min_phy_value'] = kargs.get('min_phy_value', 0)
                self['max_phy_value'] = kargs.get('max_phy_value', 0)
                for i in range(kargs['links_nr'] - 5):
                    self['val_{}'.format(i)] = kargs['val_{}'.format(i)]
                    
            elif kargs['conversion_type'] == CONVERSION_TYPE_RTABX:
                self['block_len'] = ((kargs['links_nr'] - 5) * 8 * 3) + 88
                self['links_nr'] = kargs['links_nr']
                self['name_addr'] = kargs.get('name_addr', 0)
                self['unit_addr'] = kargs.get('unit_addr', 0)
                self['comment_addr'] = kargs.get('comment_addr', 0)
                self['inv_conv_addr'] = kargs.get('inv_conv_addr', 0)
                for i in range(kargs['links_nr'] - 5):
                    self['text_{}'.format(i)] = 0
                self['default_addr'] = kargs.get('default_addr', 0)
                self['conversion_type'] = CONVERSION_TYPE_RTABX
                self['precision'] = kargs.get('precision', 0)
                self['flags'] = kargs.get('flags', 0)
                self['ref_param_nr'] = kargs['links_nr'] - 4
                self['val_param_nr'] = (kargs['links_nr'] - 5) * 2
                self['min_phy_value'] = kargs.get('min_phy_value', 0)
                self['max_phy_value'] = kargs.get('max_phy_value', 0)
                for i in range(kargs['links_nr'] - 5):
                    self['lower_{}'.format(i)] = kargs['lower_{}'.format(i)]
                    self['upper_{}'.format(i)] = kargs['upper_{}'.format(i)]

    def __bytes__(self):
        fmt = '<4sI{}Q2B3H{}d'.format(self['links_nr'] + 2, self['val_param_nr'] + 2)

        # only compute keys for Python < 3.6
        if PYVERSION_MAJOR < 36:
            if self['conversion_type'] == CONVERSION_TYPE_NON:
                keys = KEYS_CONVERSION_NONE
            elif self['conversion_type'] == CONVERSION_TYPE_LIN:
                keys = KEYS_CONVERSION_LINEAR
            elif self['conversion_type'] == CONVERSION_TYPE_RAT:
                keys = KEYS_CONVERSION_RAT
            elif self['conversion_type'] == CONVERSION_TYPE_ALG:
                keys = KEYS_CONVERSION_ALGEBRAIC
            elif self['conversion_type'] in (CONVERSION_TYPE_TABI, CONVERSION_TYPE_TAB):
                keys = KEYS_CONVERSION_NONE
                for i in range(self['val_param_nr'] // 2):
                    keys += ('raw_{}'.format(i), 'phys_{}'.format(i))
            elif self['conversion_type'] == CONVERSION_TYPE_RTAB:
                keys = KEYS_CONVERSION_NONE
                for i in range(self['val_param_nr']):
                    keys += ('lower{}'.format(i), 'upper_{}'.format(i), 'phys_{}'.format(i))
                keys += ('default',)
            elif self['conversion_type'] == CONVERSION_TYPE_TABX:
                keys = ('id',
                        'reserved0',
                        'block_len',
                        'links_nr',
                        'name_addr',
                        'unit_addr',
                        'comment_addr',
                        'inv_conv_addr')
                keys += tuple('text_{}'.format(i) for i in range(self['links_nr'] - 4 - 1))
                keys += ('default_addr',)
                keys += ('conversion_type',
                         'precision',
                         'flags',
                         'ref_param_nr',
                         'val_param_nr',
                         'min_phy_value',
                         'max_phy_value')
                keys += tuple('val_{}'.format(i) for i in range(self['val_param_nr']))
            elif self['conversion_type'] == CONVERSION_TYPE_RTABX:
                keys = ('id',
                        'reserved0',
                        'block_len',
                        'links_nr',
                        'name_addr',
                        'unit_addr',
                        'comment_addr',
                        'inv_conv_addr')
                keys += tuple('text_{}'.format(i) for i in range(self['links_nr'] - 4 - 1))
                keys += ('default_addr',)
                keys += ('conversion_type',
                         'precision',
                         'flags',
                         'ref_param_nr',
                         'val_param_nr',
                         'min_phy_value',
                         'max_phy_value')
                for i in range(self['val_param_nr'] // 2):
                    keys += ('lower_{}'.format(i), 'upper_{}'.format(i))
            elif self['conversion_type'] == CONVERSION_TYPE_TTAB:
                keys = ('id',
                        'reserved0',
                        'block_len',
                        'links_nr',
                        'name_addr',
                        'unit_addr',
                        'comment_addr',
                        'inv_conv_addr')
                keys += tuple('text_{}'.format(i) for i in range(self['links_nr'] - 4))
                keys += ('conversion_type',
                         'precision',
                         'flags',
                         'ref_param_nr',
                         'val_param_nr',
                         'min_phy_value',
                         'max_phy_value')
                keys += tuple('val_{}'.format(i) for i in range(self['val_param_nr'] -1))
                keys += ('val_default',)
            elif self['conversion_type'] == CONVERSION_TYPE_TRANS:
                keys = ('id',
                        'reserved0',
                        'block_len',
                        'links_nr',
                        'name_addr',
                        'unit_addr',
                        'comment_addr',
                        'inv_conv_addr')
                for i in range((self['links_nr'] - 4 -1) // 2):
                    keys += ('input_{}_addr'.format(i), 'output_{}_addr'.format(i))
                keys += ('default_addr',
                         'conversion_type',
                         'precision',
                         'flags',
                         'ref_param_nr',
                         'val_param_nr',
                         'min_phy_value',
                         'max_phy_value')
                keys += tuple('val_{}'.format(i) for i in range(self['val_param_nr'] -1))
                keys += ('val_default',)

        if PYVERSION_MAJOR >= 36:
            return pack(fmt, *self.values())
        else:
            return pack(fmt, *[self[key] for key in keys])


class DataBlock(dict):
    """DTBLOCK class

    Parameters
    ----------
    address : int
        DTBLOCK address inside the file
    file_stream : int
        file handle

    """
    def __init__(self, **kargs):
        super(DataBlock, self).__init__()

        try:
            self.address = address = kargs['address']
            stream = kargs['file_stream']
            stream.seek(address, SEEK_START)

            (self['id'],
             self['reserved0'],
             self['block_len'],
             self['links_nr']) = unpack(FMT_COMMON, stream.read(COMMON_SIZE))
            self['data'] = stream.read(self['block_len'] - COMMON_SIZE)

        except KeyError:

            self['id'] = b'##DT'
            self['reserved0'] = 0
            self['block_len'] = len(kargs['data']) + COMMON_SIZE
            self['links_nr'] = 0
            self['data'] = kargs['data']

    def __bytes__(self):
        fmt = FMT_DATA_BLOCK.format(self['block_len'] - COMMON_SIZE)
        if PYVERSION_MAJOR >= 36:
            return pack(fmt, *self.values())
        else:
            return pack(fmt, *[self[key] for key in KEYS_DATA_BLOCK])



class DataZippedBlock(dict):
    """DZBLOCK class

    Parameters
    ----------
    address : int
        DTBLOCK address inside the file
    file_stream : int
        file handle

    """
    def __init__(self, **kargs):
        super(DataZippedBlock, self).__init__()

        self.prevent_data_setitem = True
        try:
            self.address = address = kargs['address']
            stream = kargs['file_stream']
            stream.seek(address, SEEK_START)

            (self['id'],
             self['reserved0'],
             self['block_len'],
             self['links_nr'],
             self['original_type'],
             self['zip_type'],
             self['reserved1'],
             self['param'],
             self['original_size'],
             self['zip_size'],) = unpack(FMT_DZ_COMMON, stream.read(DZ_COMMON_SIZE))

            self['data'] = stream.read(self['zip_size'])

        except KeyError:
            self.prevent_data_setitem = False

            data = kargs['data']

            self['id'] = b'##DZ'
            self['reserved0'] = 0

            self['links_nr'] = 0
            self['original_type'] = kargs.get('original_type', 'DT')
            self['zip_type'] = karg.get('zip_type', FLAG_DZ_DEFLATE)
            self['reserved1'] = 0
            self['param'] = 0 if self['zip_type'] == FLAG_DZ_DEFLATE else kargs['param']

            # since prevent_data_setitem is False the rest of the keys will be handled by __setitem__
            self['data'] = data

        self.prevent_data_setitem = False
        self.return_unzipped = True

    def __setitem__(self, item, value):
        if item == 'data' and self.prevent_data_setitem == False:
            data = value
            self['original_size'] = len(data)

            if self['zip_type'] == FLAG_DZ_DEFLATE:
                data = compress(data)
            else:
                cols = self['param']
                lines = self['original_size'] // cols

                nd = np.fromstring(data[:lines*cols], dtype=np.uint8).reshape((lines, cols))
                data = nd.transpose().tostring() + data[lines*cols:]

                data = compress(data)

            self['zip_size'] = len(data)
            self['block_len'] = self['zip_size'] + DZ_COMMON_SIZE
            super(DataZippedBlock, self).__setitem__(item, data)
        else:
            super(DataZippedBlock, self).__setitem__(item, value)

    def __getitem__(self, item):
        if item == 'data':
            if self.return_unzipped:
                data = super(DataZippedBlock, self).__getitem__(item)
                data = decompress(data)
                if self['zip_type'] == FLAG_DZ_DEFLATE:
                    return data
                else:
                    cols = self['param']
                    lines = self['original_size'] // cols

                    nd = np.fromstring(data[:lines*cols], dtype=np.uint8).reshape((cols, lines))
                    data = nd.transpose().tostring() + data[lines*cols:]

                    return data
            else:
                return super(DataZippedBlock, self).__getitem__(item)
        else:
            return super(DataZippedBlock, self).__getitem__(item)

    def __bytes__(self):
        fmt = FMT_DZ_COMMON + '{}s'.format(self['zip_size'])
        self.return_unzipped = False
        if PYVERSION_MAJOR >= 36:
            data = pack(fmt, *self.values())
        else:
            data = pack(fmt, *[self[key] for key in KEYS_DZ_BLOCK])
        self.return_unzipped = True
        return data


class DataGroup(dict):
    """DGBLOCK class"""
    def __init__(self, **kargs):
        super(DataGroup, self).__init__()

        try:
            self.address = address = kargs['address']
            stream = kargs['file_stream']
            stream.seek(address, SEEK_START)

            (self['id'],
             self['reserved0'],
             self['block_len'],
             self['links_nr'],
             self['next_dg_addr'],
             self['first_cg_addr'],
             self['data_block_addr'],
             self['comment_addr'],
             self['record_id_len'],
             self['reserved1']) = unpack(FMT_DATA_GROUP, stream.read(DG_BLOCK_SIZE))

        except KeyError:

            self.address = 0
            self['id'] = kargs.get('id', '##DG'.encode('utf-8'))
            self['reserved0'] = kargs.get('reserved0', 0)
            self['block_len'] = kargs.get('block_len', DG_BLOCK_SIZE)
            self['links_nr'] = kargs.get('links_nr', 4)
            self['next_dg_addr'] = kargs.get('next_dg_addr', 0)
            self['first_cg_addr'] = kargs.get('first_cg_addr', 0)
            self['data_block_addr'] = kargs.get('data_block_addr', 0)
            self['comment_addr'] = kargs.get('comment_addr', 0)
            self['record_id_len'] = kargs.get('record_id_len', 0)
            self['reserved1'] = kargs.get('reserved1', b'\00'*7)

    def __bytes__(self):
        if PYVERSION_MAJOR >= 36:
            return pack(FMT_DATA_GROUP, *self.values())
        else:
            return pack(FMT_DATA_GROUP, *[self[key] for key in KEYS_DATA_GROUP])


class DataList(dict):
    """DLBLOCK class"""
    def __init__(self, **kargs):
        super(DataList, self).__init__()

        try:
            self.address = address = kargs['address']
            stream = kargs['file_stream']
            stream.seek(address, SEEK_START)

            (self['id'],
             self['reserved0'],
             self['block_len'],
             self['links_nr']) = unpack(FMT_COMMON, stream.read(COMMON_SIZE))

            self['next_dl_addr'] = unpack('<Q', stream.read(8))[0]

            links = unpack('<{}Q'.format(self['links_nr'] - 1), stream.read( (self['links_nr'] - 1) * 8 ))

            for i, addr in enumerate(links):
                self['data_block_addr{}'.format(i)] = addr

            (self['flags'],
             self['reserved1'],
             self['data_block_nr'],
             self['data_block_len']) = unpack('<B3sIQ', stream.read(16))

        except KeyError:

            self.address = 0
            self['id'] = kargs.get('id', '##DL'.encode('utf-8'))
            self['reserved0'] = kargs.get('reserved0', 0)
            self['block_len'] = kargs.get('block_len', CN_BLOCK_SIZE)
            self['links_nr'] = kargs.get('links_nr', 8)
            self['next_dl_addr'] = kargs.get('next_dl_addr', 0)

            for i in range(self['links_nr'] - 1):
                self['data_block_addr{}'.format(i)] = kargs.get('data_block_addr{}'.format(i), 0)

            self['flags'] = kargs.get('flags', 1)
            self['reserved1'] = kargs.get('reserved1', b'\00'*3)
            self['data_block_nr'] = kargs.get('data_block_nr', 1)
            self['data_block_len'] = kargs.get('data_block_len', 1)

    def __bytes__(self):
        fmt = FMT_DATA_LIST.format(self['links_nr'])
        if PYVERSION_MAJOR < 36:
            keys = ('id', 'reserved0', 'block_len', 'links_nr', 'next_dl_addr')
            keys += tuple('data_block_addr{}'.format(i) for i in range(self['links_nr'] - 1))
            keys += ('flags', 'reserved1', 'data_block_nr', 'data_block_len')
        if PYVERSION_MAJOR >= 36:
            return pack(fmt, *self.values())
        else:
            return pack(fmt, *[self[key] for key in keys])


class FileIdentificationBlock(dict):
    """IDBLOCK class"""
    def __init__(self, **kargs):

        super(FileIdentificationBlock, self).__init__()

        self.address = 0

        try:

            stream = kargs['file_stream']
            stream.seek(self.address, SEEK_START)

            (self['file_identification'],
             self['version_str'],
             self['program_identification'],
             self['reserved0'],
             self['reserved1'],
             self['mdf_version'],
             self['reserved2'],
             self['check_block'],
             self['fill'],
             self['unfinalized_standard_flags'],
             self['unfinalized_custom_flags']) = unpack(FMT_IDENTIFICATION_BLOCK, stream.read(IDENTIFICATION_BLOCK_SIZE))

        except KeyError:

            version = kargs.get('version', 400)
            self['file_identification'] = 'MDF     '.encode('utf-8')
            self['version_str'] = '{}    '.format(version).encode('utf-8')
            self['program_identification'] = 'Python  '.encode('utf-8')
            self['reserved0'] = 0
            self['reserved1'] = 0
            self['mdf_version'] = int(version.replace('.', ''))
            self['reserved2'] = 0
            self['check_block'] = 0
            self['fill'] = b'\x00' * 26
            self['unfinalized_standard_flags'] = 0
            self['unfinalized_custom_flags'] = 0

    def __bytes__(self):
        if PYVERSION_MAJOR >= 36:
            return pack(FMT_IDENTIFICATION_BLOCK, *self.values())
        else:
            return pack(FMT_IDENTIFICATION_BLOCK, *[self[key] for key in KEYS_IDENTIFICATION_BLOCK])


class FileHistory(dict):
    """FHBLOCK class"""
    def __init__(self, **kargs):
        super(FileHistory, self).__init__()

        try:
            self.address = address = kargs['address']
            stream = kargs['file_stream']
            stream.seek(address, SEEK_START)

            (self['id'],
             self['reserved0'],
             self['block_len'],
             self['links_nr'],
             self['next_fh_addr'],
             self['comment_addr'],
             self['abs_time'],
             self['tz_offset'],
             self['daylight_save_time'],
             self['time_flags'],
             self['reserved1']) = unpack(FMT_FILE_HISTORY, stream.read(FH_BLOCK_SIZE))

        except KeyError:
            self['id'] = kargs.get('id', '##FH'.encode('utf-8'))
            self['reserved0'] = kargs.get('reserved0', 0)
            self['block_len'] = kargs.get('block_len', FH_BLOCK_SIZE)
            self['links_nr'] = kargs.get('links_nr', 2)
            self['next_fh_addr'] = kargs.get('next_fh_addr', 0)
            self['comment_addr'] = kargs.get('comment_addr', 0)
            self['abs_time'] = kargs.get('abs_time', int(time.time()) * 10**9)
            self['tz_offset'] = kargs.get('tz_offset', 120)
            self['daylight_save_time'] = kargs.get('daylight_save_time', 60)
            self['time_flags'] = kargs.get('time_flags', 2)
            self['reserved1'] = kargs.get('reserved1', b'\x00'*3)

    def __bytes__(self):
        if PYVERSION_MAJOR >= 36:
            return pack(FMT_FILE_HISTORY, *self.values())
        else:
            return pack(FMT_FILE_HISTORY, *[self[key] for key in KEYS_FILE_HISTORY])


class HeaderBlock(dict):
    """HDBLOCK class"""
    def __init__(self, **kargs):
        super(HeaderBlock, self).__init__()

        try:
            self.address = address = kargs['address']
            stream = kargs['file_stream']
            stream.seek(address, SEEK_START)

            (self['id'],
             self['reserved3'],
             self['block_len'],
             self['links_nr'],
             self['first_dg_addr'],
             self['file_history_addr'],
             self['channel_tree_addr'],
             self['first_attachment_addr'],
             self['first_event_addr'],
             self['comment_addr'],
             self['abs_time'],
             self['tz_offset'],
             self['daylight_save_time'],
             self['time_flags'],
             self['time_quality'],
             self['flags'],
             self['reserved4'],
             self['start_angle'],
             self['start_distance']) = unpack(FMT_HEADER_BLOCK, stream.read(HEADER_BLOCK_SIZE))

        except KeyError:

            self['id'] = '##HD'.encode('utf-8')
            self['reserved3'] = kargs.get('reserved3' , 0)
            self['block_len'] = kargs.get('block_len' , HEADER_BLOCK_SIZE)
            self['links_nr'] = kargs.get('links_nr' , 6)
            self['first_dg_addr'] = kargs.get('first_dg_addr' , 0)
            self['file_history_addr'] = kargs.get('file_history_addr' , 0)
            self['channel_tree_addr'] = kargs.get('channel_tree_addr' , 0)
            self['first_attachment_addr'] = kargs.get('first_attachment_addr' , 0)
            self['first_event_addr'] = kargs.get('first_event_addr' , 0)
            self['comment_addr'] = kargs.get('comment_addr' , 0)
            self['abs_time'] = kargs.get('abs_time' , int(time.time()) * 10**9)
            self['tz_offset'] = kargs.get('tz_offset' , 120)
            self['daylight_save_time'] = kargs.get('daylight_save_time' , 60)
            self['time_flags'] = kargs.get('time_flags' , 2)     #offset valid
            self['time_quality'] = kargs.get('time_quality' , 0) #time source PC
            self['flags'] = kargs.get('flags' , 0)
            self['reserved4'] = kargs.get('reserved4', 0)
            self['start_angle'] = kargs.get('start_angle' , 0)
            self['start_distance'] = kargs.get('start_distance' , 0)

    def __bytes__(self):
        if PYVERSION_MAJOR >= 36:
            return pack(FMT_HEADER_BLOCK, *self.values())
        else:
            return pack(FMT_HEADER_BLOCK, *[self[key] for key in KEYS_HEADER_BLOCK])


class HeaderList(dict):
    """HLBLOCK class"""
    def __init__(self, **kargs):
        super(HeaderList, self).__init__()

        try:
            self.address = address = kargs['address']
            stream = kargs['file_stream']
            stream.seek(address, SEEK_START)

            (self['id'],
             self['reserved0'],
             self['block_len'],
             self['links_nr'],
             self['first_dl_addr'],
             self['flags'],
             self['zip_type'],
             self['reserved1']) = unpack(FMT_HL_BLOCK, stream.read(HL_BLOCK_SIZE))

        except KeyError:

            self.address = 0
            self['id'] = b'##HL'
            self['reserved0'] = 0
            self['block_len'] = HL_BLOCK_SIZE
            self['links_nr'] = 1
            self['first_dl_addr'] = 0
            self['flags'] = 1
            self['zip_type'] = 0
            self['reserved1'] = b'\x00' * 5

    def __bytes__(self):
        if PYVERSION_MAJOR >= 36:
            return pack(FMT_HL_BLOCK, *self.values())
        else:
            return pack(FMT_HL_BLOCK, *[self[key] for key in KEYS_HL_BLOCK])


class SourceInformation(dict):
    """SIBLOCK class"""
    def __init__(self, **kargs):
        super(SourceInformation, self).__init__()

        try:
            self.address = address = kargs['address']
            stream = kargs['file_stream']
            stream.seek(address, SEEK_START)

            (self['id'],
             self['reserved0'],
             self['block_len'],
             self['links_nr'],
             self['name_addr'],
             self['path_addr'],
             self['comment_addr'],
             self['source_type'],
             self['bus_type'],
             self['flags'],
             self['reserved1']) = unpack(FMT_SOURCE_INFORMATION, stream.read(SI_BLOCK_SIZE))

        except KeyError:
            self.address = 0
            self['id'] = b'##SI'
            self['reserved0'] = 0
            self['block_len'] = SI_BLOCK_SIZE
            self['links_nr'] = 3
            self['name_addr'] = 0
            self['path_addr'] = 0
            self['comment_addr'] = 0
            self['source_type'] = kargs.get('source_type', SOURCE_TOOL)
            self['bus_type'] = kargs.get('bus_type', BUS_TYPE_NONE)
            self['flags'] = 0
            self['reserved1'] = b'\x00' * 5

    def __bytes__(self):
        if PYVERSION_MAJOR >= 36:
            return pack(FMT_SOURCE_INFORMATION, *self.values())
        else:
            return pack(FMT_SOURCE_INFORMATION, *[self[key] for key in KEYS_SOURCE_INFORMATION])


class SignalDataBlock(dict):
    """SDBLOCK class"""
    def __init__(self, **kargs):
        super(SignalDataBlock, self).__init__()

        try:
            self.address = address = kargs['address']
            stream = kargs['file_stream']
            stream.seek(address, SEEK_START)

            (self['id'],
             self['reserved0'],
             self['block_len'],
             self['links_nr']) = unpack(FMT_COMMON, stream.read(COMMON_SIZE))
            self['data'] = stream.read(self['block_len'] - COMMON_SIZE)

        except KeyError:
            self.address = 0
            self['id'] = b'##SD'
            self['reserved0'] = 0
            data = kargs['data']
            self['block_len'] = len(data) + COMMON_SIZE
            self['links_nr'] = 0
            self['data'] = data

    def __bytes__(self):
        fmt = FMT_DATA_BLOCK.format(self['block_len'] - COMMON_SIZE)
        keys = KEYS_DATA_BLOCK
        if PYVERSION_MAJOR >= 36:
            res = pack(fmt, *self.values())
        else:
            res = pack(fmt, *[self[key] for key in keys])
        # 8 byte alignment
        size = len(res)
        if size % 8:
            res += b'\x00' * (8 - size%8)
        return res


class TextBlock(dict):
    """common TXBLOCK and MDBLOCK class"""
    def __init__(self, **kargs):
        super(TextBlock, self).__init__()

        try:
            stream = kargs['file_stream']
            self.address = address = kargs['address']

            stream.seek(address, SEEK_START)
            (self['id'],
             self['reserved0'],
             self['block_len'],
             self['links_nr']) = unpack(FMT_COMMON, stream.read(COMMON_SIZE))

            size = self['block_len'] - COMMON_SIZE

            self['text'] = text = stream.read(size)

            self.text_str = text.decode('utf-8').strip('\x00')

        except KeyError:

            self.address = 0
            text = kargs['text']

            try:
                self.text_str = text
                text = text.encode('utf-8')
            except:
                self.text_str = text.decode('utf-8').strip(' \n\t\x00')

            text_length = len(text)
            align = text_length % 8
            if align:
                padding = 8 - align
            elif text[-1] in (0, b'\x00'):
                padding = 0
            else:
                padding = 8

            self['id'] = kargs['id']
            self['reserved0'] = 0
            self['block_len'] = text_length + padding + COMMON_SIZE
            self['links_nr'] = 0
            if padding:
                self['text'] = text + b'\00' * padding
            else:
                self['text'] = text

    @classmethod
    def from_text(cls, text, meta=False):
        """Create a TextBlock from a str or bytes

        Parameters
        ----------
        text : str | bytes
            input text
        meta : bool
            enable meta text block

        Returns
        -------

        Examples
        --------
        >>> t = TextBlock.from_text(b'speed')
        >>> t['id']
        b'##TX'
        >>> t.text_str
        speed
        >>> t = TextBlock.from_text('mass', meta=True)
        >>> t['id']
        b'##MD'

        """
        kargs = {}

        kargs['id'] = b'##MD' if meta else b'##TX'
        kargs['text'] = text

        return cls(**kargs)

    def __bytes__(self):
        fmt = FMT_TEXT_BLOCK.format(self['block_len'] - COMMON_SIZE)
        if PYVERSION_MAJOR >= 36:
            return pack(fmt, *self.values())
        else:
            return pack(fmt, *[self[key] for key in KEYS_TEXT_BLOCK])
